#include <iostream>
#include <fstream>
#include <unistd.h> //included for the sleep function
#include <getopt.h>
#include "TROOT.h"

#include "Event.h"
#include "HecEvent.h"
#include "HecCell.h"
#include "EmecCell.h"
#include "HecRunHeader2002v1.h"
#include "HecEventNTup2002v1.h"
#include "GlobalHeader2002v0.h"
#include "EmecRunHeader2002v0.h"
#include "EmecEventNTup2002v1.h"
#include "SlowControl2002v0.h"
#include "Headers.h"
#include "Geometry.h"
#include "TApplication.h"
#include "App.h"
#include "TBrowser.h"


#include "UserAlg.h"

/*! Class:    No Class. This file contains the main() function.
  * Author:   Ian Gable (igable@uvic.ca)
  * Date:     05/10/2002
  * Description:  main() parses the command line arguments
  * Modifications:  030803 M. Lefebvre:  allow to run with no data files
  *                 030805 M. Lefebvre:  add -H -E and -g options
  */

// strip the filename from a path-and-filename string
string stripFileName( string pathAndFileName){
  string fileName("");
  string::size_type lastSlash = pathAndFileName.find_last_of("/");
  if(lastSlash < pathAndFileName.size()){
    fileName = pathAndFileName.substr(lastSlash+1,pathAndFileName.size()-1);
  } else {
    return fileName;
  }
}

// read a file that contains, in each line, the path to a data file
vector<string> readDataPathsFile(string dataPathsFileName) {
  // open the file
  ifstream fStream(dataPathsFileName.c_str());
  // exit if problems opening the file
  if(!fStream.is_open()){
    cout << "readDataPathsFile failed to open file:" << dataPathsFileName << endl;
    exit(1);
  }
  vector<string> dataPathsVector;
  while (!fStream.eof()) {
    string path = "";
    fStream >> path;
    if (!fStream.eof()) dataPathsVector.push_back(path);
  }
  fStream.close();
  cout << "Read " << dataPathsVector.size() << " data file paths from " << dataPathsFileName << endl;
  return dataPathsVector;
}

// The main control loop for the program.
// This method handles creating the necessary HecEvent, Geometry and UserAlg
// objects. It then passes control to the user at the apropriate time.
void run(string hecFileName, string emecFileName, string hecGeoFileName, string emecGeoFileName, string globalGeoFileName){

  cout << "***** Starting TBRootAna execution..." << endl;

  //hec specific
  HecEventNTup2002v1* hecEventNTuple = 0;
  HecRunHeader2002v1* hecRunHeaderNTuple = 0;
  SlowControl2002v0* hecSlowControlNTuple = 0;
  GlobalHeader2002v0* hecGlobalRunHeaderNTuple = 0;
  if (hecFileName != "") {
    //create all the ntuples found in a combined hec NTuple file
    hecEventNTuple = new HecEventNTup2002v1(0, hecFileName);
    hecRunHeaderNTuple = new HecRunHeader2002v1(0, hecFileName);
    hecSlowControlNTuple =  new SlowControl2002v0(0, hecFileName);
    hecGlobalRunHeaderNTuple = new GlobalHeader2002v0(0, hecFileName);
    // load the information
    hecRunHeaderNTuple->setDataFileName(stripFileName(hecFileName));
    hecRunHeaderNTuple->LoadTree(0);
    hecRunHeaderNTuple->fChain->GetEntry(0);
    hecSlowControlNTuple->LoadTree(0);
    hecSlowControlNTuple->fChain->GetEntry(0);
    hecGlobalRunHeaderNTuple->LoadTree(0);
    hecGlobalRunHeaderNTuple->getChain()->GetEntry(0);
    //
    cout << "hec NTuples" << endl;
    cout << "  run number: " << hecGlobalRunHeaderNTuple->getRunNo() << endl;
    cout << "  number of events: " << hecEventNTuple->numEvents() << endl;
    cout << "  febno from " << hecRunHeaderNTuple->getFirstCh() 
         << " to " << hecRunHeaderNTuple->getLastCh() << endl;
 }

  //emec specific
  EmecEventNTup2002v1* emecEventNTuple = 0;
  EmecRunHeader2002v0* emecRunHeaderNTuple = 0;
  SlowControl2002v0* emecSlowControlNTuple = 0;
  GlobalHeader2002v0* emecGlobalRunHeaderNTuple = 0;
  if (emecFileName != "") {
   //create all the ntuples found in a combined emec NTuple file
    emecEventNTuple = new EmecEventNTup2002v1(0, emecFileName);
    emecRunHeaderNTuple = new EmecRunHeader2002v0(0, emecFileName);
    emecSlowControlNTuple =  new SlowControl2002v0(0, emecFileName);
    emecGlobalRunHeaderNTuple = new GlobalHeader2002v0(0, emecFileName);
    // load the information
    emecRunHeaderNTuple->setDataFileName(stripFileName(emecFileName));
    emecRunHeaderNTuple->LoadTree(0);
    emecRunHeaderNTuple->fChain->GetEntry(0);
    emecSlowControlNTuple->LoadTree(0);
    emecSlowControlNTuple->fChain->GetEntry(0);
    emecGlobalRunHeaderNTuple->LoadTree(0);
    emecGlobalRunHeaderNTuple->getChain()->GetEntry(0);
    //
    cout << "emec NTuples" << endl;
    cout << "  run number: " << emecGlobalRunHeaderNTuple->getRunNo() << endl;
    cout << "  number of events: " << emecEventNTuple->numEvents() << endl;
    cout << "  febno from " << emecRunHeaderNTuple->getFirstCh() 
         << " to " << emecRunHeaderNTuple->getLastCh() << endl;
  }

  // check if hec and emec NTuples correspond to the same run
  if (hecFileName != "" && emecFileName != "") {
    if (hecGlobalRunHeaderNTuple->getRunNo() != emecGlobalRunHeaderNTuple->getRunNo()) {
      cout << "\n+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+\n"
           << "WARNING!!!!!!! : EMEC and HEC run numbers are not the same\n"
           << "Do you have the right files?\n\n";
    }
  }

  // choose slow control and global header ntuple (hec wins)
  SlowControl2002v0* slowControlNTuple = 0;
  GlobalHeader2002v0* globalRunHeaderNTuple = 0;
  if (hecFileName != "") {
    slowControlNTuple = hecSlowControlNTuple;
    globalRunHeaderNTuple = hecGlobalRunHeaderNTuple;
  } else if (emecFileName != "") {
    slowControlNTuple = emecSlowControlNTuple;
    globalRunHeaderNTuple = emecGlobalRunHeaderNTuple;
  }

  // add all the header objects to the headers wrapper class
  // note that some or all of the pointers could be undefined
  Headers* headers = new Headers(hecRunHeaderNTuple, emecRunHeaderNTuple, slowControlNTuple, globalRunHeaderNTuple);

  // create geometry service
  Geometry* geo = new Geometry(hecGeoFileName, emecGeoFileName, globalGeoFileName);

  // create empty event
  Event* event = new Event();

  // create UserAlg
  cout << "Creating UserAlg..." << endl;
  UserAlg* userAlg = new UserAlg(event, geo, headers);

  // user initialize
  cout << "Excuting UserAlg::initialize()..." << endl;
  userAlg->initialize();

  // find number of events to analyse
  // if ambiguous, take smallest number of events
  int numEvents = 0;
  if (hecFileName != "" && emecFileName != "") {
    numEvents = hecEventNTuple->numEvents();
    if (emecEventNTuple->numEvents() < numEvents) numEvents = emecEventNTuple->numEvents();
  } else if (hecFileName != "") {
    numEvents = hecEventNTuple->numEvents();
  } else if (emecFileName != "") {
    numEvents = emecEventNTuple->numEvents();
  }

  // loop over events
  if (hecFileName != "" || emecFileName != "") {
    cout << "Beginnning event loop..." << endl;

    for(int indexE = 0; indexE < numEvents; indexE++){
      if (indexE%1000 == 0) cout << "Number of events processed " << indexE << endl;
      
      if (hecFileName != "") {
        // hec
        // load the information
        hecEventNTuple->LoadTree(indexE);
        hecEventNTuple->fChain->GetEntry(indexE); // very important apparently
        // fill event
        event->hecEventNumber(hecEventNTuple->evtno);
        for(int febno = hecRunHeaderNTuple->getFirstCh(); febno <= hecRunHeaderNTuple->getLastCh(); febno++){
          HecCell cell(febno, hecEventNTuple->signal[febno-1]); // memory inefficient....
          event->addHecCell(cell);
        }
        // not really hec specific
        event->setHecTrigger(hecEventNTuple->trig); // this is a pointer to an array;
        event->setBeamChamberInfo(hecEventNTuple->slop, hecEventNTuple->pos0, hecEventNTuple->xcov, hecEventNTuple->ycov);
        event->setGlobalTimingInfo(hecEventNTuple->tdc_ns, hecEventNTuple->Globalcubictime, hecEventNTuple->Globalcubictimesigma);
      }

      if (emecFileName != "") {
        // emec
        // load the information
        emecEventNTuple->LoadTree(indexE);
        emecEventNTuple->fChain->GetEntry(indexE); // very important apparently
        // fill event
        event->emecEventNumber(emecEventNTuple->evtno );
        for(int febno = emecRunHeaderNTuple->getFirstCh(); febno <= emecRunHeaderNTuple->getLastCh(); febno++){
          EmecCell cell(febno, emecEventNTuple->signal[febno]); // memory inefficient...
          event->addEmecCell(cell);
        }
        // not really emec specific
        event->setEmecTrigger(emecEventNTuple->trig); // this is a pointer to an array;
        if (hecFileName == "") {  // hec wins
          event->setBeamChamberInfo(emecEventNTuple->slop, emecEventNTuple->pos0, emecEventNTuple->xcov, emecEventNTuple->ycov);
          event->setGlobalTimingInfo(emecEventNTuple->tdc_ns, emecEventNTuple->Globalcubictime, emecEventNTuple->Globalcubictimesigma);
        }
      }

      // check event numbers
      if (hecFileName != "" && emecFileName != "") {
        if(event->hecEventNumber() != event->emecEventNumber()){
          cout << "\n+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+*+\n"
               << "WARNING!!!!!!! : HEC and Emec event numbers mismatched\n"
               << "Posible HEC EMEC out of phase problem?\n\n";
        }
      }

      // user event control
      userAlg->execute();

      //  clear the cell so that new values can be loaded at next event
      event->clearHecCells();
      event->clearEmecCells();
    }
    cout << "Number of events processed " << numEvents << endl;
  }

  // user finalyse
  cout << "Executing UserAlg::finalize() ...\n";
  userAlg->finalize();

  // delete objects
  if (hecFileName != "") {
    delete hecEventNTuple;
    delete hecRunHeaderNTuple;
    delete hecSlowControlNTuple;
    delete hecGlobalRunHeaderNTuple;
  }
  if (emecFileName != "") {
    delete emecEventNTuple;
    delete emecRunHeaderNTuple;
    delete emecSlowControlNTuple;
    delete emecGlobalRunHeaderNTuple;
  }
  delete userAlg;
  delete headers;
  delete event;
  delete geo;
  cout << "Run objects deleted" << endl;
}


int  main(int argc, char * argv[]){
  // options
  // -h hecNtupleFile1 hecNtupleFile2 etc.
  // -h hecNtupleFile1 -h hecNtupleFile2 etc.
  // -e emecNtupleFile1 emecNtupleFile2 etc.
  // -e emecNtupleFile1 -e emecNtupleFile2 etc.
  // -a hecGeoFileName otherStuffIgnored
  //    default file name used if unspecified
  // -b emecGeoFileName otherStuffIgnored
  //    default file name used if unspecified
  // -g globalGeoFileName otherStuffIgnored
  //    default file name used if unspecified
  // -H file1 file2 etc.
  // -H file1 -H file2 etc.
  //    if file contains the following lines
  //    hecNtupleFile1
  //    hecNtupleFile2
  //    then -H file is equivalent to -h hecNtupleFile1 hecNtupleFile2
  // -E file1 file2 etc.
  // -E file1 -E file2 etc.
  //    if file contains the following lines
  //    emecNtupleFile1
  //    emecNtupleFile2
  //    then -E file is equivalent to -e emecNtupleFile1 emecNtupleFile2
  //
  // Options -a, -b and -g only consider the first argument.
  // For example,
  // -a file junk junk stuff
  // is equivalent to
  // -a file
  //
  // If any of options -a, -b or -g is used more than once, only the last
  // one is retained.  For example,
  // -a file1 -a file2
  // is equivalent to
  // -a file2

  TApplication myapp("myapp",0,0); //this will make the X11 libs load and allow you
                                                  // to draw objects. Fix for 3.05

  gROOT->SetApplication(&myapp);
  //myapp.Run();


  cout << "Begining Execution... \n";
  gROOT->GetListOfClasses();


  string commandArgs(*argv);
  string pathToExecutable("");
  string::size_type end =commandArgs.find_first_of(" ");
  string firstThing = commandArgs.substr(0,end);
  string::size_type lastSlash = firstThing.find_last_of("/");
  if(lastSlash < firstThing.size()){
    pathToExecutable = firstThing.substr(0,lastSlash+1);
    cout << "Path to tbrootana executable: " <<  pathToExecutable <<endl;
  }

  // parse the command line arguments
  // do not modify this code unless you're very brave!
  string hecFileName("");
  string emecFileName("");
  string hecGeoFileName("");
  string emecGeoFileName("");
  string globalGeoFileName("");
  string hecFileNamesFileName("");
  string emecFileNamesFileName("");
  
  bool isHecFile = false;
  bool isEmecFile = false;
  bool isHecGeoFile = false;
  bool isEmecGeoFile = false;
  bool isGlobalGeoFile = false;
  bool isHecFileNamesFile = false;
  bool isEmecFileNamesFile = false;

  vector<string> hecFileVector;
  vector<string> emecFileVector;
     
  if(argc > 1){
    for(int i = 0; i <argc-1;i++){
      
      char* pchar = argv[i+1];
      
      switch( pchar[0] ){
      case '-': {
        switch( pchar[1] ){
	case 'h': // hec data file name
	  isHecFile = true;       
	  isEmecFile = false;
	  isHecGeoFile = false;
	  isEmecGeoFile = false;
          isGlobalGeoFile = false;
          isHecFileNamesFile = false;
          isEmecFileNamesFile = false;
	  break;
	case 'e': // emec data file name
	  isHecFile = false;
	  isEmecFile = true;
	  isHecGeoFile = false;
	  isEmecGeoFile = false;
          isGlobalGeoFile = false;
          isHecFileNamesFile = false;
          isEmecFileNamesFile = false;
	  break;
	case 'a': // hec geo file name
	  isHecFile = false;
	  isEmecFile = false;
	  isHecGeoFile = true;
	  isEmecGeoFile = false;
          isGlobalGeoFile = false;
          isHecFileNamesFile = false;
          isEmecFileNamesFile = false;
	  break;
	case 'b': // emec geo file name
	  isHecFile = false;
	  isEmecFile = false;
	  isHecGeoFile = false;
	  isEmecGeoFile = true;
          isGlobalGeoFile = false;
          isHecFileNamesFile = false;
          isEmecFileNamesFile = false;
	  break;
	case 'g': // global geo file name
	  isHecFile = false;
	  isEmecFile = false;
	  isHecGeoFile = false;
	  isEmecGeoFile = false;
          isGlobalGeoFile = true;
          isHecFileNamesFile = false;
          isEmecFileNamesFile = false;
	  break;
	case 'H': // hec data file paths file name
	  isHecFile = false;
	  isEmecFile = false;
	  isHecGeoFile = false;
	  isEmecGeoFile = false;
          isGlobalGeoFile = false;
          isHecFileNamesFile = true;
          isEmecFileNamesFile = false;
	  break;
	case 'E': // hec data file paths file name
	  isHecFile = false;
	  isEmecFile = false;
	  isHecGeoFile = false;
	  isEmecGeoFile = false;
          isGlobalGeoFile = false;
          isHecFileNamesFile = false;
          isEmecFileNamesFile = true;
	  break;
	default:
	  cout << "Unrecognized option -" << pchar[1] << endl;
	  exit(1);
	  break;
	}
	break;
      }
      default: {
	// handle the file names
        if (isHecFile) {
	  hecFileName = pchar;
	  hecFileVector.push_back(hecFileName);
        } else if (isEmecFile) {
	  emecFileName = pchar;
	  emecFileVector.push_back(emecFileName);
        } else if (isHecGeoFile) {
	  hecGeoFileName = pchar;
          isHecGeoFile = false;
	} else if (isEmecGeoFile) {
	  emecGeoFileName = pchar;
          isHecGeoFile = false; 
	} else if (isGlobalGeoFile) {
	  globalGeoFileName = pchar;
          isGlobalGeoFile = false; 
        } else if (isHecFileNamesFile) {
	  hecFileNamesFileName = pchar;
          vector<string> fileVector = readDataPathsFile(hecFileNamesFileName);
          for (vector<string>::iterator it = fileVector.begin(); it != fileVector.end(); ++it) {
            hecFileVector.push_back(*it);
          }
        } else if (isEmecFileNamesFile) {
	  emecFileNamesFileName = pchar;
          vector<string> fileVector = readDataPathsFile(emecFileNamesFileName);
          for (vector<string>::iterator it = fileVector.begin(); it != fileVector.end(); ++it) {
            emecFileVector.push_back(*it);
          }
	} else {
	  // handle more cases
	}
	break;
      }
      }
      
    }
  }

  //check to see if we are using default geometry
  if(hecGeoFileName == ""){
    hecGeoFileName = pathToExecutable + "../geo/hec_geo_030629.dat";
    cout << "Using default hec geometry" << endl;
  }
  if(emecGeoFileName == ""){
    emecGeoFileName = pathToExecutable + "../geo/emec_geo_030801.dat";
    cout << "Using default emec geometry" << endl;
  }
  if(globalGeoFileName == ""){
    globalGeoFileName = pathToExecutable + "../geo/global_geo_030617.dat";
    cout << "Using default global geometry" << endl;
  }

  cout << "hec geometry file: " << hecGeoFileName <<endl;
  cout << "emec geometry file: " << emecGeoFileName <<endl;
  cout << "global geometry file: " << globalGeoFileName << endl;
  cout << "number of hec data files given: " << hecFileVector.size() << endl;
  cout << "number of emec data files given: " << emecFileVector.size() << endl;

  // pad the *FileVector with "" if required
  int hecNFile = hecFileVector.size();
  int emecNFile = emecFileVector.size();
  if (hecNFile == 0 && emecNFile == 0) {
    hecFileVector.push_back("");
    emecFileVector.push_back("");
  } else if (hecNFile > emecNFile) {
    for (int i = 0; i < hecNFile - emecNFile; ++i) {
      emecFileVector.push_back("");
    }
  } else if (emecNFile > hecNFile) {
    for (int i = 0; i < emecNFile - hecNFile; ++i) {
      hecFileVector.push_back("");
    }
  }

  if(hecFileVector.size() == 1){
    cout << "***** TBRootAna will run once:" << endl;
  } else {
    cout << "***** TBRootAna will run " << hecFileVector.size() << " times:" << endl;
  }
  vector<string>::iterator it1, it2;
  for (it1 = hecFileVector.begin(), it2 = emecFileVector.begin(); it1 != hecFileVector.end(); it1++, it2++){
    cout << "data files: hec: ";
    if (*it1 != "") {
      cout << *it1 << endl;
    } else {
      cout << "no hec file" << endl;
    }
    cout << "           emec: ";
    if (*it2 != "") {
      cout << *it2 << endl;
    } else {
      cout << "no emec file" << endl;
    }
  }
  
  sleep(3);// let people look at the list for three seconds
  
  for(int i = 0; i < hecFileVector.size(); i++){
    run(hecFileVector[i], emecFileVector[i], hecGeoFileName, emecGeoFileName, globalGeoFileName);
  }
  App mainWindow(gClient->GetRoot(), 400, 220);
  //TBrowser*  b = new TBrowser();
  //myapp.Run();
  
  myapp.Run();
  return 0;
}

