#include "Geometry.h"
#include "TROOT.h"
#include "TStyle.h"

#include <fstream>
#include <sstream>
#include <cmath>
#include <list>

#include "TGeoManager.h"
#include "TGeoMatrix.h"
#include "TGeoMaterial.h"
#include "TLine.h"
#include "TCanvas.h"
#include "TGeoArb8.h"
#include "TH1.h"
#include "TH2.h"
#include "TPolyLine.h"
#include "TMarker.h"
#include "TAxis.h"
#include "TText.h"
#include "TFrame.h"
#include "TArc.h"


using std::ifstream;
using std::cos;
using std::sin;
using std::sinh;
using std::ostringstream;
using std::cin;
using std::list;
using std::cerr;

Geometry::Geometry(string hecGeoFileName, string emecGeoFileName,string globalGeometryFileName){
  m_hec2DcanvasNumber = 0;
  m_emec2DcanvasNumber =0;
  m_emec2DcanvasNumberDetail =0;
  m_hec2DcanvasNumberDetail =0;

  // load geometry files
  loadHecGeometryFile(hecGeoFileName);
  loadEmecGeometryFile(emecGeoFileName);
  loadGlobalGeometryFile(globalGeometryFileName);

  // build channel id map
  buildIDMap();

  // build the hec and emac geometry maps
  hecGeo();
  emecGeo();

  buildTouchingNeighborsEmec();
  buildTouchingNeighborsHec();

  //draw3DGeometry();
  //test();

}

// ==== public hec methods =======================

int Geometry::hec_ieta(int febno){
  if(hecCheckBounds(febno, "Hec ieta accesses outside of bounds at: "))
    return (int)m_hec_ieta[febno-1];
}

int Geometry::hec_iphi(int febno){
  if(hecCheckBounds(febno, "Hec iphi accesses outside of bounds at: "))
    return (int)m_hec_iphi[febno-1];
}

int Geometry::hec_iz(int febno){
  if(hecCheckBounds(febno, "Hec iz accesses outside of bounds at: "))
    return (int)m_hec_iz[febno-1];
}

int Geometry::hec_imod(int febno){
  if(hecCheckBounds(febno, "Hec imod accesses outside of bounds at: ")) {
    int iphi = hec_iphi(febno);
    if (hec_ieta(febno)<=10) {
      if (hec_iz(febno)==3) return (int)(iphi/2);
      return (int)((iphi+1)/2);
    }
    else 
      return iphi;
  }
}

float Geometry::hec_eta(int febno){
  if(hecCheckBounds(febno, "Hec eta accesses outside of bounds at: "))
    return m_hec_eta[febno-1];
}

float Geometry::hec_phi(int febno){
  if(hecCheckBounds(febno, "Hec phi accesses outside of bounds at: "))
    return m_hec_phi[febno-1];
}

float Geometry::hec_z(int febno){
  if(hecCheckBounds(febno, "Hec z accesses outside of bounds at: "))
    return m_hec_z[febno-1];
}

float Geometry::hec_deta(int febno){
  if(hecCheckBounds(febno, "Hec deta accesses outside of bounds at: "))
    return m_hec_deta[febno-1];
}

float Geometry::hec_dphi(int febno){
  if(hecCheckBounds(febno, "Hec dphi accesses outside of bounds at: "))
    return m_hec_dphi[febno-1];
}

float Geometry::hec_dz(int febno){
  if(hecCheckBounds(febno, "Hec dz accesses outside of bounds at: "))
    return m_hec_dz[febno-1];
}

bool Geometry::hec_isConnected(int febno){
  if(febno >= m_hecFirstFebno && febno <= m_hecLastFebno){
    if(m_hec_eta[febno-1] == 0) // cannot use hec_eta(febno) because it calls hec_isConnected!
      return false;
    else
      return true;
  } else {
    cout << "Hec is connected checked out of bounds at: " << febno << endl;
    exit(1);
  }
}

map<int,TVector3> Geometry::hec_LocalCenterMap(){
  return m_hecLocalCenterMap;
}

map<int,TVector3> Geometry::hec_TBCenterMap(){
  return m_hecTBCenterMap;
}

TVector3 Geometry::hec_LocalCenter(int febno){
  if(hecCheckBounds(febno, "Hec cell Local center accessed out of bounds at: "))
    return m_hecLocalCenterMap[febno];
}

TVector3 Geometry::hec_TBCenter(int febno){
  if(hecCheckBounds(febno, "Hec cell TB center accessed out of bounds at: "))
    return m_hecTBCenterMap[febno]; 
}

map<int,TVector3> Geometry::hec_LocalGeoCenterMap(){
  return m_hecLocalGeoCenterMap;
}

map<int,TVector3> Geometry::hec_TBGeoCenterMap(){
  return m_hecTBGeoCenterMap;
}

TVector3 Geometry::hec_LocalGeoCenter(int febno){
  if(hecCheckBounds(febno, "Hec cell Local GeoCenter accessed out of bounds at: "))
    return m_hecLocalGeoCenterMap[febno];
}

TVector3 Geometry::hec_TBGeoCenter(int febno){
  if(hecCheckBounds(febno, "Hec cell TB GeoCenter accessed out of bounds at: "))
    return m_hecTBGeoCenterMap[febno];
}

double Geometry::hec_Volume(int febno){
  if(hecCheckBounds(febno, "Hec cell volume accessed out of bounds at: "))
    return m_hecVolumeMap[febno];
}

multimap< double,int > Geometry::hec_neighbors(TVector3 point){
  multimap<double,int > near;
  typedef multimap< double,int >::value_type map_type;
  // here we calculate the distance between a particular point and all the
  // centers of the cells in the hec. We then insert this distance into a multimap
  // thus allowing us to sort them in order from smallest to biggest.
  map<int,TVector3>::iterator it;
  for(it = m_hecTBCenterMap.begin(); it!=m_hecTBCenterMap.end(); it++){
    near.insert(map_type((it->second-point).Mag() ,it->first));
  }
  return near;
}

multimap< double,int > Geometry::hec_neighbors(TVector3 point, int layer){
  multimap<double,int > near;
  typedef multimap< double,int >::value_type map_type;
  map<int,TVector3>::iterator it;
  for(it = m_hecTBCenterMap.begin(); it != m_hecTBCenterMap.end(); it++){
    if(hec_iz(it->first) == layer)  // check the layer
      near.insert(map_type((it->second-point).Mag(), it->first));
  }
  return near;
}

multimap< double,int > Geometry::hec_neighbors(vector<float> lineParameters){
  multimap<double,int > near;
  typedef multimap< double,int >::value_type map_type;

  if(lineParameters.size() != 4){
    cout << "ERROR: You have provided and insuficient line parameterization" <<endl;
    exit(1);
  }

  float a0 = lineParameters[0];
  float b0 = lineParameters[1];
  float a1 = lineParameters[2];
  float b1 = lineParameters[3];

  TVector3 r0(a0, b0, 0);                        // vector to a point on the line at z=0
  TVector3 nhat = (TVector3(a1, b1, 1)).Unit();  // unit vector in the direction on the line

  map<int,TVector3>::iterator it;
  for(it = m_hecTBCenterMap.begin(); it != m_hecTBCenterMap.end(); it++){
    TVector3 d = (r0 - it->second) - ((nhat*(r0 - it->second))*nhat);
    near.insert(map_type(d.Mag(), it->first));
  }
  return near;
}

multimap< double,int > Geometry::hec_neighbors(vector<float> lineParameters, int layer){
  multimap<double,int > near;
  typedef multimap< double,int >::value_type map_type;

  if(lineParameters.size() != 4){
    cout << "ERROR: You have provided and insuficient line parameterization" <<endl;
    exit(1);
  }

  float a0 = lineParameters[0];
  float b0 = lineParameters[1];
  float a1 = lineParameters[2];
  float b1 = lineParameters[3];

  TVector3 r0(a0, b0, 0);                        // vector to a point on the line at z=0
  TVector3 nhat = (TVector3(a1, b1, 1)).Unit();  // unit vector in the direction on the line

  map<int,TVector3>::iterator it;
  for(it = m_hecTBCenterMap.begin(); it != m_hecTBCenterMap.end(); it++){
    if(hec_iz(it->first) == layer){
      TVector3 d = (r0 - it->second) - ((nhat*(r0 - it->second))*nhat);
      near.insert(map_type(d.Mag(), it->first));
    }
  }
  return near;
}

vector<int> Geometry::hec_neighbors(int febno){
  return m_hecTouch[febno];
}

vector<int> Geometry::hec_neighbors(int febno, int layer){
  vector<int> inLayer;
  for(vector<int>::iterator it = m_hecTouch[febno].begin(); it != m_hecTouch[febno].end(); it++){
    if(layer == emec_iz(*it)){
      inLayer.push_back(*it);
    }
  }
  return inLayer;
}


TVector3 Geometry::hec_LocalToTB(TVector3 local){
  return ((*m_tbHecRotation)*local)+(*m_tbHecVectorTrans);
}

TVector3 Geometry::hec_TBToLocal(TVector3 tb){
  return (m_tbHecRotation->Inverse())*(tb-(*m_tbHecVectorTrans));
}

// ==== public emec methods =======================

int Geometry::emec_ieta(int febno){
  if(emecCheckBounds(febno, "Emec ieta accesses outside of bounds at: "))
    return (int)m_emec_ieta[febno];
}

int Geometry::emec_iphi(int febno){
  if(emecCheckBounds(febno, "Emec iphi accesses outside of bounds at: "))
    return (int)m_emec_iphi[febno];
}

int Geometry::emec_iz(int febno){
  if(emecCheckBounds(febno, "Emec iz accesses outside of bounds at: "))
    return (int)m_emec_iz[febno];
}

float Geometry::emec_eta(int febno){
  if(emecCheckBounds(febno, "Emec eta accesses outside of bounds at: "))
    return m_emec_eta[febno];
}

float Geometry::emec_phi(int febno){
  if(emecCheckBounds(febno, "Emec phi accesses outside of bounds at: "))
    return m_emec_phi[febno];
}

float Geometry::emec_z(int febno){
  if(emecCheckBounds(febno, "Emec z accesses outside of bounds at: "))
    return m_emec_z[febno];
}

float Geometry::emec_deta(int febno){
  if(emecCheckBounds(febno, "Emec deta accesses outside of bounds at: "))
    return m_emec_deta[febno];
}

float Geometry::emec_dphi(int febno){
  if(emecCheckBounds(febno, "Emec dphi accesses outside of bounds at: "))
    return m_emec_dphi[febno];
}

float Geometry::emec_dz(int febno){
  if(emecCheckBounds(febno, "Emec dz accesses outside of bounds at: "))
    return m_emec_dz[febno];
}

bool Geometry::emec_isConnected(int febno){
  if(febno >= m_emecFirstFebno && febno <= m_emecLastFebno){
    if(m_emec_eta[febno] == 0) // cannot use emec_eta(febno) because it calls hec_isConnected!
      return false;
    else
      return true;
  } else {
    cout << "Emec is connected checked out of bounds at: " << febno << endl;
    exit(1);
  }
}

map<int,TVector3> Geometry::emec_LocalCenterMap(){
  return m_emecLocalCenterMap;
}

map<int,TVector3> Geometry::emec_TBCenterMap(){
  return m_emecTBCenterMap;
}

TVector3 Geometry::emec_LocalCenter(int febno){
  if(emecCheckBounds(febno, "Emec cell Local center accessed out of bounds at: "))
    return m_emecLocalCenterMap[febno];
}

TVector3 Geometry::emec_TBCenter(int febno){
  if(emecCheckBounds(febno, "Emec cell TB center accessed out of bounds at: "))
    return m_emecTBCenterMap[febno];
}

map<int,TVector3> Geometry::emec_LocalGeoCenterMap(){
  return m_emecLocalGeoCenterMap;
}

map<int,TVector3> Geometry::emec_TBGeoCenterMap(){
  return m_emecTBGeoCenterMap;
}

TVector3 Geometry::emec_LocalGeoCenter(int febno){
  if(emecCheckBounds(febno, "Emec cell Local GeoCenter accessed out of bounds at: "))
    return m_emecLocalGeoCenterMap[febno];
}

TVector3 Geometry::emec_TBGeoCenter(int febno){
  if(emecCheckBounds(febno, "Emec cell TB GeoCenter accessed out of bounds at: "))
    return m_emecTBGeoCenterMap[febno];
}

double Geometry::emec_Volume(int febno){
  if(emecCheckBounds(febno, "Emec cell volume accessed out of bounds at: "))
    return m_emecVolumeMap[febno];
}

multimap< double,int > Geometry::emec_neighbors(TVector3 point){
  multimap<double,int > near;
  typedef multimap< double,int >::value_type map_type;
  map<int,TVector3>::iterator it;
  for(it = m_emecTBCenterMap.begin(); it!=m_emecTBCenterMap.end(); it++){
    near.insert(map_type((it->second-point).Mag() ,it->first));
  }
  return near;
}

multimap< double,int > Geometry::emec_neighbors(TVector3 point, int layer){
  multimap<double,int > near;
  typedef multimap< double,int >::value_type map_type;
  map<int,TVector3>::iterator it;
  for(it = m_emecTBCenterMap.begin(); it != m_emecTBCenterMap.end(); it++){
    if(emec_iz(it->first) == layer) //check the layer
      near.insert(map_type((it->second-point).Mag(), it->first));
  }
  return near;
}

multimap< double,int > Geometry::emec_neighbors(vector<float> lineParameters){
  multimap<double,int > near;
  typedef multimap< double,int >::value_type map_type;

  if(lineParameters.size() != 4){
    cout << "ERROR: You have provided and insuficient line parameterization" <<endl;
    exit(1);
  }

  float a0 = lineParameters[0];
  float b0 = lineParameters[1];
  float a1 = lineParameters[2];
  float b1 = lineParameters[3];

  TVector3 r0(a0, b0, 0);                        // vector to a point on the line at z=0
  TVector3 nhat = (TVector3(a1, b1, 1)).Unit();  // unit vector in the direction on the line

  map<int,TVector3>::iterator it;
  for(it = m_emecTBCenterMap.begin(); it!=m_emecTBCenterMap.end(); it++){
    TVector3 d = (r0 - it->second) - ((nhat*(r0 - it->second))*nhat);
    near.insert(map_type(d.Mag(), it->first));
  }
  return near;
}

multimap< double,int > Geometry::emec_neighbors(vector<float> lineParameters, int layer){
  multimap<double,int > near;
  typedef multimap< double,int >::value_type map_type;

  if(lineParameters.size() != 4){
    cout << "ERROR: You have provided and insuficient line parameterization" <<endl;
    exit(1);
  }

  float a0 = lineParameters[0];
  float b0 = lineParameters[1];
  float a1 = lineParameters[2];
  float b1 = lineParameters[3];

  TVector3 r0(a0, b0, 0);                        // vector to a point on the line at z=0
  TVector3 nhat = (TVector3(a1, b1, 1)).Unit();  // unit vector in the direction on the line

  map<int,TVector3>::iterator it;
  for(it = m_emecTBCenterMap.begin(); it!=m_emecTBCenterMap.end(); it++){
    if(emec_iz(it->first) == layer){
      TVector3 d = (r0 - it->second) - ((nhat*(r0 - it->second))*nhat);
      near.insert(map_type(d.Mag(), it->first));
    }
  }
  return near;
}

vector<int> Geometry::emec_neighbors(int febno, int layer){
  vector<int> inLayer;
  for(vector<int>::iterator it = m_emecTouch[febno].begin(); it != m_emecTouch[febno].end(); it++){
    if(layer == emec_iz(*it)){
      inLayer.push_back(*it);
    }
  }
  return inLayer;
}

vector<int> Geometry::emec_neighbors(int febno){
  return m_emecTouch[febno];
}

TVector3 Geometry::emec_LocalToTB(TVector3 local){
  return ((*m_tbEmecRotation)*local)+(*m_tbEmecVectorTrans);
}

TVector3 Geometry::emec_TBToLocal(TVector3 tb){
  return (m_tbEmecRotation->Inverse())*(tb-(*m_tbEmecVectorTrans));
}

// ==== public pan-detector methods =============

int Geometry::FEBSerialChannelNumber(int idet, int ieta, int iphi, int iz){
  if((ieta < 0) || (iphi < 0) || (iz < 0)){
    return -1;
  }
  map< TBID , int >::iterator it = m_idMap.find(TBID(idet,ieta,iphi,iz));
  if(it == m_idMap.end()){
    return -1;
  } else {
    return it->second;
  }
}

int Geometry::FEBSerialChannelNumber(TBID id){
  map< TBID , int >::iterator it = m_idMap.find(id);
  if(it == m_idMap.end()){
    return -1;
  } else {
    return it->second;
  }
}

TBID Geometry::getTBID(int idet, int febno){
  if(idet == 2){
    return TBID(idet,hec_ieta(febno),hec_iphi(febno),hec_iz(febno));
  } else if(idet == 1){
    return TBID(idet,emec_ieta(febno),emec_iphi(febno),emec_iz(febno));
  } else {
    cerr << "Get TBID was called with an invalid idet : " << idet << endl;
    exit(1);
  }
}

// ==== other public methods =============

TVector3 Geometry::epzTOxyz(double eta, double phi, double z){
  double ro = z/sinh(eta);
  double x = ro*cos(phi);
  double y = ro*sin(phi);
  return TVector3(x,y,z);
}

TVector3 Geometry::rpzTOxyz(TVector3 rhoPhi){
  double x = rhoPhi.x()*cos(rhoPhi.y());
  double y = rhoPhi.x()*sin(rhoPhi.y());
  return TVector3(x,y,rhoPhi.z());
}

//====== hec private methods =======================

void Geometry::hecGeo(){

  // To get the TRotation class to do a passive rotation simply take the negative angles
  m_tbHecRotation = new TRotation();

  // Build a passive root rotation matrix with our particular set of Y Euler Angles
  // This means that we are using the Y-convention rotation.
  m_tbHecRotation->SetYEulerAngles(-m_tbHecPhiEuler,-m_tbHecThetaEuler,-m_tbHecPsiEuler);

  // We create a vector to translate the coordinates after doing the rotation
  m_tbHecVectorTrans = new TVector3(m_tbHecXTrans,m_tbHecYTrans,m_tbHecZTrans);

  // get the min and max rho of the hec
  // get the inner hec rho in the first layer
  int febno_rho = FEBSerialChannelNumber(2, 14, 1, 1); // iz=1 and ieta=14
  m_hecInnerRhoMin = hec_z(febno_rho)/sinh(hec_eta(febno_rho) + hec_deta(febno_rho)/2);
  // get the inner hec rho in all but the first layer
  febno_rho = FEBSerialChannelNumber(2, 13, 1, 2); // iz=2 and ieta=13
  m_hecInnerRhoMax = hec_z(febno_rho)/sinh(hec_eta(febno_rho) + hec_deta(febno_rho)/2);
  // get the outer hec rho
  febno_rho = FEBSerialChannelNumber(2, 1, 1, 1); // iz=1 and ieta=1
  m_hecOuterRho = hec_z(febno_rho)/sinh(hec_eta(febno_rho) - hec_deta(febno_rho)/2);
//   cout << "m_hecInnerRhoMin = " << m_hecInnerRhoMin << endl;
//   cout << "m_hecInnerRhoMax = " << m_hecInnerRhoMax << endl;
//   cout << "m_hecOuterRho    = " << m_hecOuterRho    << endl;

  // loop over cells to fill the geometry maps
  for (int febno = m_hecFirstFebno; febno <= m_hecLastFebno; febno++) {
    if(hec_isConnected(febno)){
      // fill the (median) center x,y,z (local) map
      double rho = hec_z(febno)/sinh(hec_eta(febno));
      double x = rho*cos(hec_phi(febno));
      double y = rho*sin(hec_phi(febno));
      m_hecLocalCenterMap[febno] = (TVector3(x, y, hec_z(febno)));

      // fill the (median) center x,y,z (test beam) map
      // This line does the rotation and the translation
      m_hecTBCenterMap[febno] = (((*m_tbHecRotation)*m_hecLocalCenterMap[febno])+(*m_tbHecVectorTrans));

      // fill the GeoCenter x,y,z (local) map
      m_hecLocalGeoCenterMap[febno] = hecLocalGeoCenter(febno);

      // fill the GeoCenter x,y,z (test beam) map
      // This line does the rotation and the translation
      m_hecTBGeoCenterMap[febno] = (((*m_tbHecRotation)*m_hecLocalGeoCenterMap[febno])+(*m_tbHecVectorTrans));      

      // fill the volume map
      m_hecVolumeMap[febno] = hecVolume(febno);
    }
  }
}

void Geometry::hecNotConnected(int febno){
  cout  << "You tried to get geometry information from the unconnected Hec cell: " << febno
        << "\nSince this operation makes no sense your program has been terminated "
        << "\nPlease use the method Geometry::hec_isConnected Before asking for "
        << "\ngeometry  information about a particular cell\n";
  exit(1);
}

bool Geometry::hecCheckBounds(int febno,string msg){
  if(febno >= m_hecFirstFebno && febno <= m_hecLastFebno){
    if(hec_isConnected(febno))
      return true;
    else
      hecNotConnected(febno);
  } else {
    cout << msg << febno << endl;
    exit(1);
  }
}

int Geometry::hecNumberOfFamilies(int febno){
  int iz   = hec_iz(febno);
  int ieta = hec_ieta(febno);
  if ( iz == 1 || (iz == 2 && ieta == 1) || (iz == 3 && ieta == 2) ) {
    return 1;
  } else {
    return 2;
  }
}

double Geometry::hecFamilyZ(int febno, int family){
  if (hecNumberOfFamilies(febno) == 1) {
    return hec_z(febno);
  } else {
    if (family == 1) {
      return hec_z(febno) - hec_dz(febno)/4;
    } else {
      return hec_z(febno) + hec_dz(febno)/4;
    }
  }
}

double Geometry::hecFamilyRhoMin(int febno, int family){
  int iz   = hec_iz(febno);
  int ieta = hec_ieta(febno);
  if (iz == 1 && ieta == 14) {
    return m_hecInnerRhoMin;
  } else if ((iz == 2 && ieta == 13) || (iz == 3 && ieta == 13 && family == 1)) {
    return m_hecInnerRhoMax;
  } else {
    return hecFamilyZ(febno, family)/sinh(hec_eta(febno) + hec_deta(febno)/2);
  }
}

double Geometry::hecFamilyRhoMax(int febno, int family){
  int iz   = hec_iz(febno);
  int ieta = hec_ieta(febno);
  if ((iz == 1 && ieta == 1) || 
      (iz == 2 && ieta == 1) || 
      (iz == 2 && ieta == 2 && family != 1) ||
      (iz == 3 && ieta == 2) ||
      (iz == 3 && ieta == 3 && family != 1)) {
    return m_hecOuterRho;
  } else {
    return hecFamilyZ(febno, family)/sinh(hec_eta(febno) - hec_deta(febno)/2);
  }
}

double Geometry::hecFamilyRho(int febno, int family){
  double rhomin = hecFamilyRhoMin(febno, family);
  double rhomax = hecFamilyRhoMax(febno, family);
  double g = rhomin/rhomax;
  double rho = (2./3.) * (rhomin + rhomax)*(1 + g + g*g) / ((1 + g)*(1 + g));
  // correct for normal (half module phi width) cells
  if (hec_ieta(febno) < 11) rho *= sqrt(1 + pow(tan(hec_dphi(febno)),2)/4);
  return rho;
}

double Geometry::hecFamilyVolume(int febno, int family){
  double dz = hec_dz(febno)/hecNumberOfFamilies(febno);
  double v = dz*(pow(hecFamilyRhoMax(febno, family),2) - pow(hecFamilyRhoMin(febno, family),2));
  if (hec_ieta(febno) < 11) { // normal (half module phi width) cells
    v *= tan(hec_dphi(febno))/2;
  } else {                    // full module phi width cells
    v *= tan(hec_dphi(febno)/2);
  }
  return v;
}

TVector3 Geometry::hecLocalGeoCenter(int febno){
  // get the GeoCenter phi (local)
  double phic = hec_phi(febno);
  if (hec_ieta(febno) < 11) { // normal (half module phi width) cells
    double dphic = atan(tan(hec_dphi(febno))/2);
    if ((hec_iphi(febno)%2 == 0 && hec_iz(febno) < 3) ||
        (hec_iphi(febno)%2 != 0 && hec_iz(febno) == 3)) { // "left" cells
      phic += (dphic - hec_dphi(febno)/2);
    } else {                                              // "right" cells
      phic -= (dphic - hec_dphi(febno)/2);
    }
  }
  
  // Compute the GeoCenter x,y,z (local)
  if (hecNumberOfFamilies(febno) == 1) { // cells with one family
    double rhoc = hecFamilyRho(febno, 1);
        return (TVector3(rhoc*cos(phic), rhoc*sin(phic), hecFamilyZ(febno, 1)));
  } else {                               // cells with two families
    double v1 = hecFamilyVolume(febno, 1);
    double v2 = hecFamilyVolume(febno, 2);
    double w1 = 1/(1 + v2/v1);  // the weight for family 1
    double w2 = 1 - w1;         // the weight for family 2
    double rho1c = hecFamilyRho(febno, 1);
    double rho2c = hecFamilyRho(febno, 2);
        return w1*(TVector3(rho1c*cos(phic), rho1c*sin(phic), hecFamilyZ(febno, 1))) +
               w2*(TVector3(rho2c*cos(phic), rho2c*sin(phic), hecFamilyZ(febno, 2)));
  }
}

double Geometry::hecVolume(int febno){
  if (hecNumberOfFamilies(febno) == 1) { // cells with one family
    return hecFamilyVolume(febno, 1);
  } else {                               // cells with two families
    return hecFamilyVolume(febno, 1) + hecFamilyVolume(febno, 2);
  }
}

void Geometry::buildTouchingNeighborsHec(){
  typedef map< int, vector<int> >::value_type map_type;

  for(int i = m_hecFirstFebno; i <= m_hecLastFebno; i++){
    if(hec_isConnected(i)){
      vector<int> dummy;
      m_hecTouch.insert(map_type(i, dummy));
    }
  }

  // per layer
  for(int i = m_hecFirstFebno; i <= m_hecLastFebno; i++){
    if(hec_isConnected(i)){
      int iz = hec_iz(i);
      int ieta = hec_ieta(i);
      int iphi = hec_iphi(i);
      if((ieta < 10) || (ieta > 11)){ // regular cells
        addRange(2,ieta-1,ieta+1,-1+iphi,1+iphi,iz,i);
      } else if(ieta == 11) { // fat boundary cells
        addRange(2,ieta,ieta+1,iphi-1,1+iphi,iz,i);
        if(iz < 3){ // layer 1 and 2 fat boundary cells
          addRange(2,ieta-1,ieta-1,iphi*2-2,iphi*2+1,iz,i);
        } else { // fat boundary cells in the third layer
          addRange(2,ieta-1,ieta-1,iphi*2-1,iphi*2+2,iz,i);
        }
      } else if(ieta == 10){
        addRange(2,ieta-1,ieta,iphi-1,1+iphi,iz,i);
      }
    } // if isConnected
  }// for
  // 3D

  for(int i = m_hecFirstFebno; i <= m_hecLastFebno; i++){
    if(hec_isConnected(i)){
      int iz = hec_iz(i);
      int ieta = hec_ieta(i);
      int iphi = hec_iphi(i);
      if(iz == 1){
        if(ieta != 10){
          addRange(2,ieta,ieta+1,iphi-1,iphi+1,2,i);
        } else {
          addRange(2,ieta,ieta,iphi-1,iphi+1,2,i);
          addRange(2,ieta+1,ieta+1,iphi/2,iphi/2+1,2,i);
        }
      } else if(iz == 2){
        if( (ieta>1) && (ieta<10) ){
          addRange(2,ieta,ieta+1,iphi-1,iphi+1,3,i);
        } else if(ieta == 10) {
          addRange(2,ieta,ieta,iphi-1,iphi+1,3,i);
          addRange(2,ieta+1,ieta+1,(iphi-1)/2,(iphi-1)/2 +1,3,i);
        } else if (ieta >10){
          addRange(2,ieta,ieta+1,iphi-1,iphi,3,i);
        }
     }
   }
  }
}

//===== emec private methods =====================

void Geometry::emecGeo(){

  m_tbEmecRotation = new TRotation();
  m_tbEmecRotation->SetYEulerAngles(-m_tbEmecPhiEuler,-m_tbEmecThetaEuler,-m_tbEmecPsiEuler);
  m_tbEmecVectorTrans = new TVector3(m_tbEmecXTrans,m_tbEmecYTrans,m_tbEmecZTrans);

  for (int febno = m_emecFirstFebno; febno <= m_emecLastFebno;  febno++) {
    if(emec_isConnected(febno)){
      // fill the (median) center x,y,z (local) map
      double rho = emec_z(febno)/sinh(emec_eta(febno));
      double x = rho*cos(emec_phi(febno));
      double y = rho*sin(emec_phi(febno));
      m_emecLocalCenterMap[febno] = (TVector3(x,y,emec_z(febno)));

      // fill the (median) center x,y,z (test beam) map
      // This line does the rotation and the translation
      m_emecTBCenterMap[febno] = (((*m_tbEmecRotation)*m_emecLocalCenterMap[febno])+(*m_tbEmecVectorTrans));

      // fill the GeoCenter x,y,z (local) map
      m_emecLocalGeoCenterMap[febno] = emecLocalGeoCenter(febno);

      // fill the GeoCenter x,y,z (test beam) map
      // This line does the rotation and the translation
      m_emecTBGeoCenterMap[febno] = (((*m_tbEmecRotation)*m_emecLocalGeoCenterMap[febno])+(*m_tbEmecVectorTrans));      

      // fill the volume map
      m_emecVolumeMap[febno] = emecVolume(febno);
    }
  }

}

void Geometry::emecNotConnected(int febno){
  cout  << "You tried to get geometry information from the unconnected  Emec cell: " << febno
        << "\nSince this operation makes no sense your program has been terminated "
        << "\nPlease use the method Geometry::emec_isConnected Before asking for "
        << "\ngeometry  information about a particular cell\n";
  exit(1);
}

bool Geometry::emecCheckBounds(int febno, string msg){
  if(febno >= m_emecFirstFebno && febno <= m_emecLastFebno){
      if(emec_isConnected(febno))
        return true;
      else
        emecNotConnected(febno);
  } else {
    cout << msg << febno << endl;
    exit(1);
  }
}

TVector3 Geometry::emecLocalGeoCenter(int febno){
  double s = sin(emec_dphi(febno)/2)/(emec_dphi(febno)/2);
  if (emec_iz(febno) == 0) { // presampler
    double rho1 = emec_z(febno) / sinh(emec_eta(febno) + emec_deta(febno)/2);
    double rho2 = emec_z(febno) / sinh(emec_eta(febno) - emec_deta(febno)/2);
    double rhoc = s * (2./3.) * (rho1*rho1 + rho1*rho2 + rho2*rho2)/(rho1 + rho2);
    return (TVector3(rhoc*cos(emec_phi(febno)), rhoc*sin(emec_phi(febno)), emec_z(febno)));
  } else {
    double rhoa1 = (emec_z(febno) - emec_dz(febno)/2) / sinh(emec_eta(febno) + emec_deta(febno)/2);
    double rhoa2 = (emec_z(febno) - emec_dz(febno)/2) / sinh(emec_eta(febno) - emec_deta(febno)/2);
    double rhob1 = (emec_z(febno) + emec_dz(febno)/2) / sinh(emec_eta(febno) + emec_deta(febno)/2);
    double rhob2 = (emec_z(febno) + emec_dz(febno)/2) / sinh(emec_eta(febno) - emec_deta(febno)/2);
    double d = (rhoa2*rhoa2 + rhoa2*rhob2 + rhob2*rhob2) - (rhoa1*rhoa1 + rhoa1*rhob1 + rhob1*rhob1);
    double rhoc = s * (1./2.) * ((rhoa2*rhoa2+rhob2*rhob2)*(rhoa2+rhob2) - (rhoa1*rhoa1+rhob1*rhob1)*(rhoa1+rhob1)) / d;
    double zc = emec_z(febno) * (3./2.) * ((rhob2*rhob2+rhoa2*rhoa2) - (rhoa1*rhoa1+rhob1*rhob1)) / d;
    return (TVector3(rhoc*cos(emec_phi(febno)), rhoc*sin(emec_phi(febno)), zc));
  }
}

double Geometry::emecVolume(int febno){
  if (emec_iz(febno) == 0) { // presampler
    double rho1 = emec_z(febno) / sinh(emec_eta(febno) + emec_deta(febno)/2);
    double rho2 = emec_z(febno) / sinh(emec_eta(febno) - emec_deta(febno)/2);
    return (emec_dphi(febno)/2)*emec_dz(febno)*(rho2*rho2 - rho1*rho1);
  } else {                   // emec layer 1, 2, 3
    double rhoa1 = (emec_z(febno) - emec_dz(febno)/2) / sinh(emec_eta(febno) + emec_deta(febno)/2);
    double rhoa2 = (emec_z(febno) - emec_dz(febno)/2) / sinh(emec_eta(febno) - emec_deta(febno)/2);
    double rhob1 = (emec_z(febno) + emec_dz(febno)/2) / sinh(emec_eta(febno) + emec_deta(febno)/2);
    double rhob2 = (emec_z(febno) + emec_dz(febno)/2) / sinh(emec_eta(febno) - emec_deta(febno)/2);
    return (emec_dphi(febno)/6)*emec_dz(febno)*
      ((rhoa2*rhoa2 + rhoa2*rhob2 + rhob2*rhob2) - (rhoa1*rhoa1 + rhoa1*rhob1 + rhob1*rhob1));
  }
}

void Geometry::buildTouchingNeighborsEmec(){
  typedef map< int, vector<int> >::value_type map_type;

  // per layer
  for(int i = m_emecFirstFebno; i <= m_emecLastFebno; i++){
    if(emec_isConnected(i)){
      vector<int> dummy;
      m_emecTouch.insert(map_type(i, dummy));
      addRange(1,emec_ieta(i)-1,emec_ieta(i)+1,emec_iphi(i)-1,emec_iphi(i)+1,emec_iz(i),i);
    }
  }
  // 3D
  for(int i = m_emecFirstFebno; i <= m_emecLastFebno; i++){
    if(emec_isConnected(i)){
      int iz = emec_iz(i);
      int ieta = emec_ieta(i);
      int iphi = emec_iphi(i);
      if(iz == 0){
          addRange(1, 3+8*ieta,12+8*ieta,-1+iphi,1+iphi,1,i);
      } else if(iz == 1){
        if(ieta < 4){
          addRange(1,-1+ieta,1+ieta,-1+4*iphi,4+4*iphi,2,i);
        } else {
          addRange(1,4+(ieta-4)/8,4+(ieta-4)/8,-1+4*iphi,4+4*iphi,2,i);
          if( ((ieta-3)%8) == 0){
           addRange(1,5+ (ieta-4)/8,5+ (ieta-4)/8,-1+4*iphi,4+4*iphi,2,i);
          } else if( ((ieta-4)%8) == 0){
           addRange(1,3+(ieta-4)/8,3+(ieta-4)/8,-1+4*iphi,4+4*iphi,2,i);
          }
        }
      } else if(iz == 3){
        addRange(1,3+2*ieta,6+2*ieta,-1+iphi,1+iphi,2,i);
      }
   }
  }
}

//===== pan-detector private methods

void Geometry::buildIDMap(){
  // loop over hec cells to fill the id map (hec is idet=2)
  for (int febno = m_hecFirstFebno; febno <= m_hecLastFebno; febno++) {
    if(hec_isConnected(febno)){
      m_idMap[TBID(2, hec_ieta(febno), hec_iphi(febno), hec_iz(febno))] = febno;
    }
  }
  // loop over emec cells to fill the id map (emec is idet=1)
  for (int febno = m_emecFirstFebno; febno <= m_emecLastFebno;  febno++) {
    if(emec_isConnected(febno)){
      m_idMap[TBID(1, emec_ieta(febno), emec_iphi(febno), emec_iz(febno))] = febno;
    }
  }
}

void Geometry::addRange(int detectorType,int etaLow, int etaHigh, int phiLow, int phiHigh, int z, int febno){
  for(int i = etaLow; i <= etaHigh; i++){
    for(int j = phiLow; j <= phiHigh; j++){
      if((FEBSerialChannelNumber(detectorType,i,j,z) >=0) && (FEBSerialChannelNumber(detectorType,i,j,z) != febno)){
          addTouch(detectorType, febno, FEBSerialChannelNumber(detectorType,i,j,z));
      }
    }
  }
}

void Geometry::addTouch(int detectorType, int source, int target){
  if(detectorType == 1){
    bool notContained = true;
    for(vector<int>::iterator it = m_emecTouch[source].begin(); it != m_emecTouch[source].end(); it++){
      if((*it) == target){
        notContained = false;
      }
    }
    if(notContained){
      m_emecTouch[source].push_back(target);
    }
    notContained = true;
    for(vector<int>::iterator it = m_emecTouch[target].begin(); it != m_emecTouch[target].end(); it++){
      if((*it) == source){
        notContained = false;
      }
    }
    if(notContained){
      m_emecTouch[target].push_back(source);
    }
  }
  if(detectorType == 2){
    bool notContained = true;
    for(vector<int>::iterator it = m_hecTouch[source].begin(); it != m_hecTouch[source].end(); it++){
      if((*it) == target){
        notContained = false;
      }
    }
    if(notContained){
      m_hecTouch[source].push_back(target);
    }
    notContained = true;
    for(vector<int>::iterator it = m_hecTouch[target].begin(); it != m_hecTouch[target].end(); it++){
      if((*it) == source){
        notContained = false;
      }
    }
    if(notContained){
      m_hecTouch[target].push_back(source);
    }
  }
}


//======= Drawing methods =========================

void Geometry::draw3DGeometry(){
  //--- define some materials
  TCanvas* can = new TCanvas("3dCanvas","3dCanvas",800,800);
  TGeoMaterial *matVacuum = new TGeoMaterial("Vacuum", 0,0,0);
  TGeoMaterial *matAl = new TGeoMaterial("Al", 26.98,13,2.7);

  TGeoMedium *Vacuum = new  TGeoMedium("Med1", 1, matVacuum);
   TGeoMedium *Al = new  TGeoMedium("Med2", 2, matAl);
  TGeoManager *geom = new TGeoManager("simple1", "Simple geometry");
  TGeoVolume *top = geom->MakeBox("TOP", Vacuum, 700.0, 700.0, -700.0);
  geom->SetTopVolume(top); // mandatory !


  TGeoVolume *detectorBox = geom->MakeBox("HEC1", Vacuum, 200.0, 200.0,-300.0);
  detectorBox->SetVisibility(kFALSE); // this will hold the whole detector


  for(map<int,TVector3>::iterator it = m_hecTBCenterMap.begin();it != m_hecTBCenterMap.end();it++){

    ostringstream os;
    os << "HEC" << m_hec_adc[it->first-1];
    TGeoVolume *cell = geom->MakeArb8(os.str().c_str(), Al, m_hec_dz[it->first-1]/2);

    // make the different layers different colors.
    switch(m_hec_iz[it->first-1]){
      case 1:
        cell->SetLineColor(kRed);
        break;
      case 2:
        cell->SetLineColor(kGreen);
        break;
      case 3:
        cell->SetLineColor(kBlue);
        break;
    }

    TGeoArb8 *arb = (TGeoArb8*)cell->GetShape();

    // the various corners of the volumes
    double zb = m_hec_z[it->first-1] + m_hec_dz[it->first-1]/2;
    double zf = m_hec_z[it->first-1] - m_hec_dz[it->first-1]/2;

    double et = m_hec_eta[it->first-1] + m_hec_deta[it->first-1]/2;
    double eb = m_hec_eta[it->first-1] - m_hec_deta[it->first-1]/2;

    double pl = m_hec_phi[it->first-1] + m_hec_dphi[it->first-1]/2;
    double pr = m_hec_phi[it->first-1] - m_hec_dphi[it->first-1]/2;

    TVector3 v0 = hec_LocalToTB(epzTOxyz(eb,pr,zb));
    TVector3 v1 = hec_LocalToTB(epzTOxyz(eb,pl,zb));
    TVector3 v2 = hec_LocalToTB(epzTOxyz(et,pl,zb));
    TVector3 v3 = hec_LocalToTB(epzTOxyz(et,pr,zb));
    TVector3 v4 = hec_LocalToTB(epzTOxyz(eb,pr,zf));
    TVector3 v5 = hec_LocalToTB(epzTOxyz(eb,pl,zf));
    TVector3 v6 = hec_LocalToTB(epzTOxyz(et,pl,zf));
    TVector3 v7 = hec_LocalToTB(epzTOxyz(et,pr,zf));

    arb->SetVertex(0, v0.x(), v0.y());
    arb->SetVertex(1, v1.x(), v1.y());
    arb->SetVertex(2, v2.x(), v2.y());
    arb->SetVertex(3, v3.x(), v3.y());
    arb->SetVertex(4, v4.x(), v4.y());
    arb->SetVertex(5, v5.x(), v5.y());
    arb->SetVertex(6, v6.x(), v6.y());
    arb->SetVertex(7, v7.x(), v7.y());


    detectorBox->AddNode(cell,1, new TGeoTranslation(0, 0, m_hecTBCenterMap[it->first].z()));
  }

  for(map<int,TVector3>::iterator it = m_emecTBCenterMap.begin();it != m_emecTBCenterMap.end();it++){

    // make sure they all cells have intelligent names
    ostringstream os;
    os << "EMEC" << m_emec_adc[it->first];
    TGeoVolume *cell = geom->MakeArb8(os.str().c_str(), Vacuum, m_emec_dz[it->first]/2);

    // make the different layers different colors.
    switch(m_emec_iz[it->first]){
      case 1:
        cell->SetLineColor(kRed);
        break;
      case 2:
        cell->SetLineColor(kGreen);
        break;
      case 3:
        cell->SetLineColor(kBlue);
        break;
      case 4:
        cell->SetLineColor(kYellow);
        break;
    }

    TGeoArb8 *arb = (TGeoArb8*)cell->GetShape();

    // the various corners of the volumes
    double zb = m_emec_z[it->first] + m_emec_dz[it->first]/2;
    double zf = m_emec_z[it->first] - m_emec_dz[it->first]/2;

    double et = m_emec_eta[it->first] + m_emec_deta[it->first]/2;
    double eb = m_emec_eta[it->first] - m_emec_deta[it->first]/2;

    double pl = m_emec_phi[it->first] + m_emec_dphi[it->first]/2;
    double pr = m_emec_phi[it->first] - m_emec_dphi[it->first]/2;

    TVector3 v0 = emec_LocalToTB(epzTOxyz(eb,pr,zb));
    TVector3 v1 = emec_LocalToTB(epzTOxyz(eb,pl,zb));
    TVector3 v2 = emec_LocalToTB(epzTOxyz(et,pl,zb));
    TVector3 v3 = emec_LocalToTB(epzTOxyz(et,pr,zb));
    TVector3 v4 = emec_LocalToTB(epzTOxyz(eb,pr,zf));
    TVector3 v5 = emec_LocalToTB(epzTOxyz(eb,pl,zf));
    TVector3 v6 = emec_LocalToTB(epzTOxyz(et,pl,zf));
    TVector3 v7 = emec_LocalToTB(epzTOxyz(et,pr,zf));

    arb->SetVertex(0, v0.x(), v0.y());
    arb->SetVertex(1, v1.x(), v1.y());
    arb->SetVertex(2, v2.x(), v2.y());
    arb->SetVertex(3, v3.x(), v3.y());
    arb->SetVertex(4, v4.x(), v4.y());
    arb->SetVertex(5, v5.x(), v5.y());
    arb->SetVertex(6, v6.x(), v6.y());
    arb->SetVertex(7, v7.x(), v7.y());

    // when we add the node we translate it to its z center.
    detectorBox->AddNode(cell,1, new TGeoTranslation(0, 0, m_emecTBCenterMap[it->first].z()));
  }

  //make the detector box part of the top geometry with no translation
  top->AddNode(detectorBox,1,gGeoIdentity);

  // close the geometry
  geom->CloseGeometry();

  // make things visible
  geom->SetVisOption(0);
  geom->SetVisLevel(4);
  top->Draw();
  if (gPad) gPad->x3d("e"); // spawn a 3D viever.
  //drawJunk();
}

vector<TVector2> Geometry::hecCalculate4Corners(int febno){
  vector<TVector3> rhoPhi;
  vector<TVector2> xy;

  double z = 0;
  switch(hec_iz(febno)){
    case 1:
        z = hec_z(1);
        break;
    case 2:
        z = hec_z(18) - (hec_dz(18)/4);
        break;
    case 3:
        z = hec_z(35) - (hec_dz(35)/4);
        break;
   }

  double rho1 = z/sinh(hec_eta(febno) + hec_deta(febno)/2);
  double rho2 = z/sinh(hec_eta(febno) - hec_deta(febno)/2);

  if (hec_ieta(febno) < 11) { // normal (half module phi width) cells
    if ((hec_iphi(febno)%2 == 0 && hec_iz(febno) < 3) || (hec_iphi(febno)%2 != 0 && hec_iz(febno) == 3)) { // "left" cells
      rhoPhi.push_back(TVector3(rho1,hec_phi(febno)-hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho2,hec_phi(febno)-hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho2*(1/cos(hec_dphi(febno))),hec_phi(febno)+hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho1*(1/cos(hec_dphi(febno))),hec_phi(febno)+hec_dphi(febno)/2,z));
    } else { // "right" cells
      rhoPhi.push_back(TVector3(rho1,hec_phi(febno)+hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho2,hec_phi(febno)+hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho2*(1/cos(hec_dphi(febno))),hec_phi(febno)-hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho1*(1/cos(hec_dphi(febno))),hec_phi(febno)-hec_dphi(febno)/2,z));
    }
  } else { // "fat" cells
      rhoPhi.push_back(TVector3(rho1*(1/cos(hec_dphi(febno)/2)),hec_phi(febno)+hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho1*(1/cos(hec_dphi(febno)/2)),hec_phi(febno)-hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho2*(1/cos(hec_dphi(febno)/2)),hec_phi(febno)-hec_dphi(febno)/2,z));
      rhoPhi.push_back(TVector3(rho2*(1/cos(hec_dphi(febno)/2)),hec_phi(febno)+hec_dphi(febno)/2,z));
  }

  for(vector<TVector3>::iterator it = rhoPhi.begin(); it != rhoPhi.end(); it++){
    xy.push_back(hec_LocalToTB(rpzTOxyz(*it)).XYvector());
  }
  return xy;
}

void Geometry::draw2DHec(map<int,float> inputMap, int canID, bool lockScale, float scaleLower, float scaleUpper, string userTitle){

  if (canID==0) canID=m_hec2DcanvasNumber;
  ostringstream os;
  os << "2dHec_" << canID;

  bool isOutsideScale = true;
  if((scaleUpper == 0) && (scaleLower == 0)) isOutsideScale = false;

  TCanvas * tmpCan = (TCanvas*) gROOT->FindObject(os.str().c_str());
  if (0==tmpCan) {
    cout << "draw2DHec : Create new Canvas with ID= " << canID << endl;
    ostringstream os3;
    os3 << "2D HEC Canvas " << userTitle;
    tmpCan = new TCanvas(os.str().c_str(),os3.str().c_str(),1152,384);
    tmpCan->Divide(3,1); // 3 pads for 3 hec layers
  }
  TCanvas * can = tmpCan;
  
  gStyle->SetOptStat(0);
  
  map< int,TH2F* > histoMap;
  for(int i = 1; i < 4; i++){
    ostringstream os1;
    os1 << "hecLayer" << i << "Color_" << canID;
    TH2F * existHisto = (TH2F*) gROOT->FindObject(os1.str().c_str());
    if (0==existHisto) {
      if (i==1) cout << "draw2DHec : Create new Histo with ID= " << canID << endl;
      histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-100,100,400,-50,150);
      histoMap[i]->GetXaxis()->SetTitle("x (cm)");
      histoMap[i]->GetYaxis()->SetTitle("y (cm)");
      histoMap[i]->GetZaxis()->SetLabelOffset(-0.07);
    }
    else {
      histoMap[i] = existHisto;
      histoMap[i]->Reset();
    }
  }

  // fill the histogram so that we can see the color palette and get the minimum
  // and maximum values.
  map<int, list<float> > scaleMap;

  for(map<int,float>::iterator it = inputMap.begin(); it != inputMap.end();it++){
    histoMap[hec_iz(it->first)]->Fill(m_hecTBCenterMap[it->first].x(),m_hecTBCenterMap[it->first].y(),it->second);
    scaleMap[hec_iz(it->first)].push_back(it->second);
  }

  // if we are not using an outside scale and we want to compute
  // the maximum and minimum accross all the histograms and lock the scale.

  list<float> maxVector;
  list<float> minVector;
  for(int i = 1; i < 4; i++){
    scaleMap[i].sort();
    if(scaleMap[i].size() > 0){
      maxVector.push_back(*(scaleMap[i].rbegin()));
      minVector.push_back(*(scaleMap[i].begin()));
    }
    if(*(scaleMap[i].rbegin()) == *(scaleMap[i].begin())){
      if(*(scaleMap[i].rbegin()) > 0){
        histoMap[i]->SetMaximum(*(scaleMap[i].rbegin()));
        histoMap[i]->SetMinimum(0);
      } else {
        histoMap[i]->SetMinimum(*(scaleMap[i].begin()));
        histoMap[i]->SetMaximum(0);
      }
    } else {
      histoMap[i]->SetMaximum(*(scaleMap[i].rbegin()));
      histoMap[i]->SetMinimum(*(scaleMap[i].begin()));
    }
  }

  if(!isOutsideScale && lockScale){
    maxVector.sort();
    minVector.sort();
    scaleUpper = *(maxVector.rbegin());
    scaleLower = *(minVector.begin());
  }


  for(int i = 1; i < 4; i++){
    can->cd(i);
    ostringstream os;
    os << "HEC Layer " << i;

    gPad->SetBorderMode(0);
    //gPad->SetBorderSize(1);
    gPad->Range(-100,-50,100,150);

    TText* title = new TText(0,-40, os.str().c_str());
    title->SetTextAlign(22);
    title->SetTextSize(0.04);
    if(lockScale || isOutsideScale){
      histoMap[i]->SetMaximum(scaleUpper);
      histoMap[i]->SetMinimum(scaleLower);
    }
    histoMap[i]->Draw("colz");
    gPad->GetFrame()->SetBorderMode(0);
    title->Draw();
  }

  map<int,TPolyLine*> polyMap; //contains pointers to the poly lines so that we can fill
                                             //them later

  // Draw the pad artwork
  for(map<int,TVector3>::iterator it = m_hecTBCenterMap.begin();it != m_hecTBCenterMap.end();it++){
    can->cd(hec_iz(it->first));

    vector<TVector2> fC = hecCalculate4Corners(it->first); // get the four corners that have been corrected for

    double x[5] = {fC[0].X(),fC[1].X(),fC[2].X(),fC[3].X(),fC[0].X()};
    double y[5] = {fC[0].Y(),fC[1].Y(),fC[2].Y(),fC[3].Y(),fC[0].Y()};

    TPolyLine* line = new TPolyLine(5,x,y);
    line->SetLineWidth(1);
    polyMap[it->first] = line;

    line->SetFillColor(0);
    line->Draw("f"); // draws the fill
    line->Draw();// draws the line
  }


  // The loop below fills all the 'hit cells' with the contents of the
  // input map.
  gStyle->SetPalette(1); // set the 'pretty' color spectrum palette
  for(map<int,float>::iterator it = inputMap.begin(); it != inputMap.end();it++){
    can->cd(hec_iz(it->first));
    double maxValue = histoMap[hec_iz(it->first)]->GetMaximum();
    double minValue = histoMap[hec_iz(it->first)]->GetMinimum();
    int fillColor = (int) floor(((it->second-minValue)/(maxValue-minValue))*50) +51;
    if(fillColor > 100) fillColor = 100;
    polyMap[it->first]->SetFillColor(fillColor);
    polyMap[it->first]->Draw("f");// F for fill.
    polyMap[it->first]->Draw();// draw the lines again
  }

  // draw the cells number in each of the cells.
  // this must happen last because. we want it to be on top of the fill.
  for(map<int,TVector3>::iterator it = m_hecTBCenterMap.begin();it != m_hecTBCenterMap.end();it++){
    can->cd(hec_iz(it->first));
    ostringstream os;
    os << it->first;
    TText* text = new TText(it->second.x(),it->second.y(), os.str().c_str());
    text->SetTextAlign(22);
    text->SetTextSize(0.02);
    text->Draw();
  }

  for(int i =1; i < 4 ; i++){
    can->cd(i);
    TArc* arc = new TArc(0.0,0.0, 30.0,0,360);
    arc->Draw();
  }
  m_hec2DcanvasNumber++; // must be the last thing we do
  
  can->Update();
}

void Geometry::draw2DHecDetail(map<int,float> inputMap, int canID, bool lockScale, float scaleLower, float scaleUpper, string userTitle){

  if (canID==0) canID=m_hec2DcanvasNumberDetail;

  bool isOutsideScale = true;
  if((scaleUpper == 0) && (scaleLower == 0)) isOutsideScale = false;

  map<int,TCanvas*> canMap;
  for(int i = 1; i < 4; i++){
    ostringstream os;
    os << "2dHecDetail_L"<< i << "_" << canID;
    ostringstream os2;
    os2 << "Hec L" << i << " " << userTitle;

    TCanvas * tmpCan = (TCanvas*) gROOT->FindObject(os.str().c_str());
    if (0==tmpCan) {
      cout << "draw2DHec : Create new Canvas with ID= " << canID << endl;
      tmpCan = new TCanvas(os.str().c_str(),os2.str().c_str(),750,750);
      tmpCan->Divide(3,1); // 3 pads for 3 hec layers
    }
    TCanvas * can = tmpCan;
    canMap[i] = can;
  }

  gStyle->SetOptStat(0);

  map< int,TH2F* > histoMap;
  for(int i = 1; i < 4; i++){
    ostringstream os1;
    os1 << "hecLayer" << i << "Color_" << canID;
    TH2F * existHisto = (TH2F*) gROOT->FindObject(os1.str().c_str());
    if (0==existHisto) {
      if (i==1) cout << "draw2DHec : Create new Histo with ID= " << canID << endl;
      histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-100,100,400,-50,150);
      histoMap[i]->GetXaxis()->SetTitle("x (cm)");
      histoMap[i]->GetYaxis()->SetTitle("y (cm)");
      histoMap[i]->GetZaxis()->SetLabelOffset(-0.07);
    }
    else {
      histoMap[i] = existHisto;
      histoMap[i]->Reset();
    }
  }

  // fill the histogram so that we can see the color palette and get the minimum
  // and maximum values.

  map<int, list<float> > scaleMap;

  for(map<int,float>::iterator it = inputMap.begin(); it != inputMap.end();it++){
    histoMap[hec_iz(it->first)]->Fill(m_hecTBCenterMap[it->first].x(),m_hecTBCenterMap[it->first].y(),it->second);
    scaleMap[hec_iz(it->first)].push_back(it->second);
  }

  // if we are not using an outside scale and we want to compute
  // the maximum and minimum accross all the histograms and lock the scale.
  list<float> maxVector;
  list<float> minVector;
  for(int i = 1; i < 4; i++){
    scaleMap[i].sort();
    if(scaleMap[i].size() > 0){
      maxVector.push_back(*(scaleMap[i].rbegin()));
      minVector.push_back(*(scaleMap[i].begin()));
    }
    if(*(scaleMap[i].rbegin()) == *(scaleMap[i].begin())){
      if(*(scaleMap[i].rbegin()) > 0){
        histoMap[i]->SetMaximum(*(scaleMap[i].rbegin()));
        histoMap[i]->SetMinimum(0);
      } else {
        histoMap[i]->SetMinimum(*(scaleMap[i].begin()));
        histoMap[i]->SetMaximum(0);
      }
    } else {
      histoMap[i]->SetMaximum(*(scaleMap[i].rbegin()));
      histoMap[i]->SetMinimum(*(scaleMap[i].begin()));
    }
  }

  if(!isOutsideScale && lockScale){
    maxVector.sort();
    minVector.sort();
    scaleUpper = *(maxVector.rbegin());
    scaleLower = *(minVector.begin());
  }


  for(int i = 1; i < 4; i++){
    gPad = canMap[i];
    ostringstream os;
    os << "HEC Layer " << i;

    gPad->SetBorderMode(0);
    //gPad->SetBorderSize(1);
    gPad->Range(-100,-50,100,150);

    TText* title = new TText(0,-40, os.str().c_str());
    title->SetTextAlign(22);
    title->SetTextSize(0.04);
    if(lockScale || isOutsideScale){
      histoMap[i]->SetMaximum(scaleUpper);
      histoMap[i]->SetMinimum(scaleLower);
    }
    histoMap[i]->Draw("colz");
    gPad->GetFrame()->SetBorderMode(0);
    title->Draw();
  }

  map<int,TPolyLine*> polyMap; //contains pointers to the poly lines so that we can fill
                                             //them later

  // Draw the pad artwork
  for(map<int,TVector3>::iterator it = m_hecTBCenterMap.begin();it != m_hecTBCenterMap.end();it++){
    gPad = canMap[hec_iz(it->first)];

    vector<TVector2> fC = hecCalculate4Corners(it->first); // get the four corners that have been corrected for

    double x[5] = {fC[0].X(),fC[1].X(),fC[2].X(),fC[3].X(),fC[0].X()};
    double y[5] = {fC[0].Y(),fC[1].Y(),fC[2].Y(),fC[3].Y(),fC[0].Y()};

    TPolyLine* line = new TPolyLine(5,x,y);
    line->SetLineWidth(1);
    polyMap[it->first] = line;

    line->SetFillColor(0);
    line->Draw("f"); // draws the fill
    line->Draw();// draws the line
  }


  // The loop below fills all the 'hit cells' with the contents of the
  // input map.
  gStyle->SetPalette(1); // set the 'pretty' color spectrum palette
  for(map<int,float>::iterator it = inputMap.begin(); it != inputMap.end();it++){
    gPad = canMap[hec_iz(it->first)];
    double maxValue = histoMap[hec_iz(it->first)]->GetMaximum();
    double minValue = histoMap[hec_iz(it->first)]->GetMinimum();
    int fillColor = (int) floor(((it->second-minValue)/(maxValue-minValue))*50) +51;
    if(fillColor > 100) fillColor = 100;
    polyMap[it->first]->SetFillColor(fillColor);
    polyMap[it->first]->Draw("f");// F for fill.

  }

  // draw the cells number in each of the cells.
  // this must happen last because. we want it to be on top of the fill.
  for(map<int,TVector3>::iterator it = m_hecTBCenterMap.begin();it != m_hecTBCenterMap.end();it++){
    gPad = canMap[hec_iz(it->first)];
    ostringstream os;
    os << it->first;
    TText* text = new TText(it->second.x(),it->second.y(), os.str().c_str());
    text->SetTextAlign(22);
    text->SetTextSize(0.02);
    text->Draw();
  }
  m_hec2DcanvasNumberDetail++; // must be the last thing we do
  for(map<int,TCanvas*>::iterator it = canMap.begin(); it != canMap.end();it++){
    TArc* arc  = new TArc(0.0,0.0,30.0,0,360);
    gPad = it->second;
    arc->Draw();
    it->second->Update();
  }
}

void Geometry::draw2DEmec(map<int,float> inputMap, int canID, bool lockScale, float scaleLower, float scaleUpper,string userTitle){
  map<int,TVector2> textCenterMap;
  if (canID==0) canID=m_emec2DcanvasNumber;
  ostringstream os;
  os << "2dEmec_" << canID;

  bool isOutsideScale = true;
  if((scaleUpper == 0) && (scaleLower == 0)) isOutsideScale = false;
  
  TCanvas * tmpCan = (TCanvas*) gROOT->FindObject(os.str().c_str());
  if (0==tmpCan) {
    cout << "draw2DEmec : Create new Canvas with ID= " << canID << endl;
    ostringstream os3;
    os3 << "Emec All Layers " << userTitle;
    tmpCan = new TCanvas(os.str().c_str(),os3.str().c_str(),1000,750);
    tmpCan->Divide(2,2); // 4 pads for 4 Emec layers
  }
  TCanvas * can = tmpCan;

  gStyle->SetOptStat(0);

  map< int,TH2F* > histoMap;
  for(int i = 0; i < 4; i++){
    ostringstream os1;
    os1 << "emecLayer" << i << "Color_" << canID;
    TH2F * existHisto = (TH2F*) gROOT->FindObject(os1.str().c_str());
    if (0==existHisto) {
      if (i==0) cout << "draw2DEmec : Create new Histo with ID= " << canID << endl;

      switch(i){
        case 0:
          histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-34.5,34.5,400,-1.0,56.0);
          break;
        case 1:
          histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-79.0,79.0, 400,-28.0,39.8);
          break;
        case 2:
          histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-75.0,75.0,400,-44.0,53.0);
          break;
        case 3:
          histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-79.0,79.0,400,-31.0,45.0);
          break;
      }
      //histoMap[i] = new TH2F(os1.str().c_str(),"",400,-100,100,400,-50,150);
      histoMap[i]->GetXaxis()->SetTitle("x (cm)");
      histoMap[i]->GetYaxis()->SetTitle("y (cm)");
      histoMap[i]->GetZaxis()->SetLabelOffset(-0.07);
    }
    else {
      histoMap[i] = existHisto;
      histoMap[i]->Reset();
    }
  }

  map<int, list<float> > scaleMap;

  for(map<int,float>::iterator it = inputMap.begin(); it != inputMap.end();it++){
    histoMap[emec_iz(it->first)]->Fill(m_emecTBCenterMap[it->first].x(),m_emecTBCenterMap[it->first].y(),it->second);
    scaleMap[emec_iz(it->first)].push_back(it->second);
  }

  // if we are not using an outside scale and we want to compute
  // the maximum and minimum accross all the histograms and lock the scale.
  list<float> maxVector;
  list<float> minVector;
  for(int i = 0; i < 4; i++){
    scaleMap[i].sort();
    if(scaleMap[i].size() > 0){
      maxVector.push_back(*(scaleMap[i].rbegin()));
      minVector.push_back(*(scaleMap[i].begin()));
    }
    if(*(scaleMap[i].rbegin()) == *(scaleMap[i].begin())){
      if(*(scaleMap[i].rbegin()) > 0){
        histoMap[i]->SetMaximum(*(scaleMap[i].rbegin()));
        histoMap[i]->SetMinimum(0);
      } else {
        histoMap[i]->SetMinimum(*(scaleMap[i].begin()));
        histoMap[i]->SetMaximum(0);
      }
    } else {
      histoMap[i]->SetMaximum(*(scaleMap[i].rbegin()));
      histoMap[i]->SetMinimum(*(scaleMap[i].begin()));
    }
  }

  if(!isOutsideScale && lockScale){
    maxVector.sort();
    minVector.sort();
    scaleUpper = *(maxVector.rbegin());
    scaleLower = *(minVector.begin());
  }


  for(int i = 1; i < 5; i++){
    can->cd(i);
    ostringstream os;
    os << "EMEC Layer " << i-1;

    gPad->SetBorderMode(0);
    //gPad->SetBorderSize(1);
    TText* title;
    switch(i){
      case 1:
        gPad->Range(-34.5,-1.0,34.5,56.0);
         title = new TText(0,52, os.str().c_str());
        break;
      case 2:
        gPad->Range(-79.0,-28.0,79.0,39.8);
        title = new TText(0,36, os.str().c_str());
        break;
      case 3:
        gPad->Range(-75.0,-44.0,75.0,53.0);
        title = new TText(0,46, os.str().c_str());
        break;
      case 4:
        gPad->Range(-79.0,-31.0,79.0,45.0);
        title = new TText(0,38, os.str().c_str());
        break;
    }

    title->SetTextAlign(22);
    title->SetTextSize(0.04);
    if(lockScale || isOutsideScale){
      histoMap[i-1]->SetMaximum(scaleUpper);
      histoMap[i-1]->SetMinimum(scaleLower);
    }
    histoMap[i-1]->Draw("colz");
    gPad->GetFrame()->SetBorderMode(0);
    title->Draw();

  }

  map<int,TPolyLine*> polyMap; //contains pointers to the poly lines so that we can fill
                                             //them later

  // Draw the pad artwork
  for(map<int,TVector3>::iterator it = m_emecTBCenterMap.begin();it != m_emecTBCenterMap.end();it++){
    can->cd(emec_iz(it->first)+1);
    double z = 0;
    switch(emec_iz(it->first)){	
       case 0:
         z = 363.0;
         break;
       case 1:
         z = 374.2;
         break;
       case 2:
         z = 395.0;
         break;
       case 3:
         z = 416.8;
         break;
     }

    double et = emec_eta(it->first) + emec_deta(it->first)/2;
    double eb = emec_eta(it->first) - emec_deta(it->first)/2;

    double pl = emec_phi(it->first) + emec_dphi(it->first)/2;
    double pr = emec_phi(it->first) - emec_dphi(it->first)/2;

    TVector2 v0 = emec_LocalToTB(epzTOxyz(eb,pr,z)).XYvector();
    TVector2 v1 = emec_LocalToTB(epzTOxyz(eb,pl,z)).XYvector();
    TVector2 v2 = emec_LocalToTB(epzTOxyz(et,pl,z)).XYvector();
    TVector2 v3 = emec_LocalToTB(epzTOxyz(et,pr,z)).XYvector();

    textCenterMap[it->first] = emec_LocalToTB(epzTOxyz(emec_eta(it->first),emec_phi(it->first),z)).XYvector();

    double x[5] = {v0.X(),v1.X(),v2.X(),v3.X(),v0.X()};
    double y[5] = {v0.Y(),v1.Y(),v2.Y(),v3.Y(),v0.Y()};

    TPolyLine* line = new TPolyLine(5,x,y);
    line->SetLineWidth(1);
    polyMap[it->first] = line;

    line->SetFillColor(0);
    line->Draw("f");
    line->Draw();
  }


  // The loop below fills all the 'hit cells' with the contents of the
  // input map.
  gStyle->SetPalette(1); // set the 'pretty' color spectrum palette
  for(map<int,float>::iterator it = inputMap.begin(); it != inputMap.end();it++){
    if(emec_isConnected(it->first)){
      can->cd(emec_iz(it->first)+1);
      double maxValue = histoMap[emec_iz(it->first)]->GetMaximum();
      double minValue = histoMap[emec_iz(it->first)]->GetMinimum();
      int fillColor = (int) floor(((it->second-minValue)/(maxValue-minValue))*50) +51;
      if(fillColor > 100) fillColor = 100;
      polyMap[it->first]->SetFillColor(fillColor);
      polyMap[it->first]->Draw("f");// F for fill.
      polyMap[it->first]->Draw();
    }
  }

  // draw the cells number in each of the cells.
  // this must happen last because. we want it to be on top of the fill.
  for(map<int,TVector2>::iterator it = textCenterMap.begin();it != textCenterMap.end();it++){
    can->cd(emec_iz(it->first)+1);
    ostringstream os;
    os << it->first;
    TText* text = new TText(it->second.X(),it->second.Y(), os.str().c_str());
    text->SetTextAlign(22);
    text->SetTextSize(0.02);
    if(emec_deta(it->first) <= emec_deta(1052)){;};
     // text->Draw();
  }
  for(int i =0; i < 4 ; i++){
    can->cd(i+1);
    TArc* arc;
    if(i == 0){
      arc = new TArc(0.0,0.0,30.0,6.5,173.5);
    } else {
      arc  = new TArc(0.0,0.0,30.0,0,360);
    }
    arc->Draw("only");
    arc->Draw();
  }

  m_emec2DcanvasNumber++; // must be the last thing we do
  can->Update();
}

void Geometry::draw2DEmecDetail(map<int,float> inputMap, int canID, bool lockScale, float scaleLower, float scaleUpper, string userTitle){
  map<int,TVector2> textCenterMap;
  if (canID==0) canID=m_emec2DcanvasNumberDetail;

  bool isOutsideScale = true;
  if((scaleUpper == 0) && (scaleLower == 0)) isOutsideScale = false;

  map<int,TCanvas*> canMap;
  for(int i = 0; i < 4; i++){
    ostringstream os;
    os << "2dEmecDetail_L"<< i << "_" << canID;
    ostringstream os2;
    os2 << "Emec L" << i << " " << userTitle;

    TCanvas * tmpCan = (TCanvas*) gROOT->FindObject(os.str().c_str());
    if (0==tmpCan) {
      cout << "draw2DHec : Create new Canvas with Layer " << i << " ID= " << canID << endl;
      tmpCan = new TCanvas(os.str().c_str(),os2.str().c_str(),1000,750);
      tmpCan->Divide(2,2); // 4 pads for 4 Emec layers
    }
    TCanvas * can = tmpCan;
    canMap[i+1] = can;
  }

  gStyle->SetOptStat(0);

  map< int,TH2F* > histoMap;
  for(int i = 0; i < 4; i++){
    ostringstream os1;
    os1 << "emecLayer" << i << "DetailColor_" << canID;
    TH2F * existHisto = (TH2F*) gROOT->FindObject(os1.str().c_str());
    if (0==existHisto) {
      if (i==0) cout << "draw2DEmec : Create new Histo with ID= " << canID << endl;

      switch(i){
        case 0:
          histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-34.5,34.5,400,-1.0,56.0);
          break;
        case 1:
          histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-79.0,79.0, 400,-28.0,39.8);
          break;
        case 2:
          histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-75.0,75.0,400,-44.0,53.0);
          break;
        case 3:
          histoMap[i] = new TH2F(os1.str().c_str(),userTitle.c_str(),400,-79.0,79.0,400,-31.0,45.0);
          break;
      }
      histoMap[i]->GetXaxis()->SetTitle("x (cm)");
      histoMap[i]->GetYaxis()->SetTitle("y (cm)");
      histoMap[i]->GetZaxis()->SetLabelOffset(-0.07);
    }
    else {
      histoMap[i] = existHisto;
      histoMap[i]->Reset();
    }
  }

  map<int, list<float> > scaleMap;
  for(map<int,float>::iterator it = inputMap.begin(); it != inputMap.end();it++){
    histoMap[emec_iz(it->first)]->Fill(m_emecTBCenterMap[it->first].x(),m_emecTBCenterMap[it->first].y(),it->second);
    scaleMap[emec_iz(it->first)].push_back(it->second);
  }
  // if we are not using an outside scale and we want to compute
  // the maximum and minimum accross all the histograms and lock the scale.
  list<float> maxVector;
  list<float> minVector;
  for(int i = 0; i < 4; i++){
    scaleMap[i].sort();
    if(scaleMap[i].size() > 0){
      maxVector.push_back(*(scaleMap[i].rbegin()));
      minVector.push_back(*(scaleMap[i].begin()));
    }
    if(*(scaleMap[i].rbegin()) == *(scaleMap[i].begin())){
      if(*(scaleMap[i].rbegin()) > 0){
        histoMap[i]->SetMaximum(*(scaleMap[i].rbegin()));
        histoMap[i]->SetMinimum(0);
      } else {
        histoMap[i]->SetMinimum(*(scaleMap[i].begin()));
        histoMap[i]->SetMaximum(0);
      }
    } else {
      histoMap[i]->SetMaximum(*(scaleMap[i].rbegin()));
      histoMap[i]->SetMinimum(*(scaleMap[i].begin()));
    }
  }

  if(!isOutsideScale && lockScale){
    maxVector.sort();
    minVector.sort();
    scaleUpper = *(maxVector.rbegin());
    scaleLower = *(minVector.begin());
  }

  for(int i = 1; i < 5; i++){
    gPad = canMap[i];
    ostringstream os;
    os << "EMEC Layer " << i-1;

    gPad->SetBorderMode(0);
    //gPad->SetBorderSize(1);
    TText* title;
    switch(i){
      case 1:
        gPad->Range(-34.5,-1.0,34.5,56.0);
         title = new TText(0,52, os.str().c_str());
        break;
      case 2:
        gPad->Range(-79.0,-28.0,79.0,39.8);
        title = new TText(0,36, os.str().c_str());
        break;
      case 3:
        gPad->Range(-75.0,-44.0,75.0,53.0);
        title = new TText(0,46, os.str().c_str());
        break;
      case 4:
        gPad->Range(-79.0,-31.0,79.0,45.0);
        title = new TText(0,38, os.str().c_str());
        break;
    }

    title->SetTextAlign(22);
    title->SetTextSize(0.04);
    if(lockScale || isOutsideScale){
      histoMap[i-1]->SetMaximum(scaleUpper);
      histoMap[i-1]->SetMinimum(scaleLower);
    }
    histoMap[i-1]->Draw("colz");
    gPad->GetFrame()->SetBorderMode(0);
    title->Draw();

  }

  map<int,TPolyLine*> polyMap; //contains pointers to the poly lines so that we can fill
                                             //them later

  // Draw the pad artwork
  for(map<int,TVector3>::iterator it = m_emecTBCenterMap.begin();it != m_emecTBCenterMap.end();it++){
    gPad = canMap[emec_iz(it->first)+1];
    double z = 0;
    switch(emec_iz(it->first)){	
       case 0:
         z = 363.0;
         break;
       case 1:
         z = 374.2;
         break;
       case 2:
         z = 395.0;
         break;
       case 3:
         z = 416.8;
         break;
     }

    double et = emec_eta(it->first) + emec_deta(it->first)/2;
    double eb = emec_eta(it->first) - emec_deta(it->first)/2;

    double pl = emec_phi(it->first) + emec_dphi(it->first)/2;
    double pr = emec_phi(it->first) - emec_dphi(it->first)/2;

    TVector2 v0 = emec_LocalToTB(epzTOxyz(eb,pr,z)).XYvector();
    TVector2 v1 = emec_LocalToTB(epzTOxyz(eb,pl,z)).XYvector();
    TVector2 v2 = emec_LocalToTB(epzTOxyz(et,pl,z)).XYvector();
    TVector2 v3 = emec_LocalToTB(epzTOxyz(et,pr,z)).XYvector();

    textCenterMap[it->first] = emec_LocalToTB(epzTOxyz(emec_eta(it->first),emec_phi(it->first),z)).XYvector();

    double x[5] = {v0.X(),v1.X(),v2.X(),v3.X(),v0.X()};
    double y[5] = {v0.Y(),v1.Y(),v2.Y(),v3.Y(),v0.Y()};

    TPolyLine* line = new TPolyLine(5,x,y);
    line->SetLineWidth(1);
    polyMap[it->first] = line;

    line->SetFillColor(0);
    line->Draw("f");
    line->Draw();
  }


  // The loop below fills all the 'hit cells' with the contents of the
  // input map.
  gStyle->SetPalette(1); // set the 'pretty' color spectrum palette
  for(map<int,float>::iterator it = inputMap.begin(); it != inputMap.end();it++){
    if(emec_isConnected(it->first)){
      gPad = canMap[emec_iz(it->first)+1];
      double maxValue = histoMap[emec_iz(it->first)]->GetMaximum();
      double minValue = histoMap[emec_iz(it->first)]->GetMinimum();
      int fillColor = (int) floor(((it->second-minValue)/(maxValue-minValue))*50) +51;
      if(fillColor > 100) fillColor = 100;
      polyMap[it->first]->SetFillColor(fillColor);
      polyMap[it->first]->Draw("f");// F for fill.
      polyMap[it->first]->Draw();
    }
  }

  // draw the cells number in each of the cells.
  // this must happen last because. we want it to be on top of the fill.
  for(map<int,TVector2>::iterator it = textCenterMap.begin();it != textCenterMap.end();it++){
    gPad = canMap[emec_iz(it->first)+1];
    ostringstream os;
    os << it->first;
    TText* text = new TText(it->second.X(),it->second.Y(), os.str().c_str());
    text->SetTextAlign(22);
    text->SetTextSize(0.012);
    text->Draw();
  }

  m_emec2DcanvasNumberDetail++; // must be the last thing we do
  for(map<int,TCanvas*>::iterator it = canMap.begin(); it != canMap.end();it++){
    TArc* arc;
    if(it->first == 1){
      arc = new TArc(0.0,0.0,30.0,6.5,173.5);
    } else {
      arc  = new TArc(0.0,0.0,30.0,0,360);
    }
    gPad = it->second;
    arc->Draw("only");
    it->second->Update();
  }
}

//=========File loading methods =====================

void Geometry::loadHecGeometryFile(string geoFileName)
{

  ifstream fStream(geoFileName.c_str());// open the file stream

  // exit of we have the wrong file name
  if(!fStream.is_open()){
    cout << "Failed to open file:" << geoFileName << endl;
    exit(1);
  }

 //parse the file
  int i = 0;
  char firstChar[2]; //first 3 Characters
  char garbage[200]; //holds discarded lines
  int junk;          //holds discarded numbers
  string firstString;//used to make comparison easier

  fStream.get(firstChar,2);// Get the first characters

  while(!fStream.eof() ){
    firstString = firstChar;

    //dump all comment lines else process them
    if(firstString == "#" ){
      fStream.getline(garbage,1000);
      fStream.get(firstChar,2);
    } else {
      fStream >> m_hec_adc[i] >> m_hec_ieta[i] >> m_hec_iphi[i] >> m_hec_iz[i]
                 >> m_hec_eta[i] >> m_hec_phi[i] >> m_hec_z[i] >> m_hec_deta[i] >> m_hec_dphi[i] >>  m_hec_dz[i];
      if (!fStream.eof()) i++;
    }
  }

  m_numHecCells = i; // numbered starting from 1
  m_hecFirstFebno = 1;
  m_hecLastFebno = m_numHecCells;
  cout << "Geometry service: number of hec febno = " << m_numHecCells
       << ", from " << m_hecFirstFebno 
       << " to " << m_hecLastFebno << endl;
  fStream.close();
}

void Geometry::loadEmecGeometryFile(string geoFileName)
{

  ifstream fStream(geoFileName.c_str());// open the file stream

  // exit of we have the wrong file name
  if(!fStream.is_open()){
    cout << "Failed to open file:" << geoFileName << endl;
    exit(1);
  }

  //parse the file
  int i = 0;
  char firstChar[2]; //first 3 Characters
  char garbage[200]; //holds discarded lines
  float junkFloat;          //holds discarded numbers
  string junkString; //

  string firstString;//used to make comparison easier

  fStream.get(firstChar,2);// Get the first characters

  while(!fStream.eof() ){
    firstString = firstChar;

    //dump all comment lines else process them
    if(firstString == "#" ){
      fStream.getline(garbage,1000);
      fStream.get(firstChar,2);
    } else {
      fStream >> m_emec_adc[i] >> m_emec_ieta[i] >> m_emec_iphi[i] >> m_emec_iz[i] >> m_emec_eta[i]
                  >> m_emec_phi[i] >> m_emec_z[i] >> m_emec_deta[i] >> m_emec_dphi[i] >>  m_emec_dz[i]
                  >> junkFloat >> junkFloat >> junkFloat >> junkString >> junkString >> junkString >> junkFloat;
      if (!fStream.eof()) i++;
    }
  }

  m_numEmecCells = i; // numbered starting from 0
  m_emecFirstFebno = 0;
  m_emecLastFebno = m_numEmecCells - 1;
  cout << "Geometry service: number of emec febno = " << m_numEmecCells
       << ", from " << m_emecFirstFebno 
       << " to " << m_emecLastFebno << endl;
  fStream.close();
}

void Geometry::loadGlobalGeometryFile(string geoFileName){
  ifstream fStream(geoFileName.c_str());// open the file stream

  // exit of we have the wrong file name
  if(!fStream.is_open()){
    cout << "Failed to open file:" << geoFileName << endl;
    exit(1);
  }

 //parse the file
  int i = 0;
  char firstChar[2]; //first 3 Characters
  char garbage[200]; //holds discarded lines
  float junkFloat;          //holds discarded numbers
  string junkString; //
  bool doneAlready= false; //used to make sure that we only read two lines of the file

  string firstString;//used to make comparison easier

  fStream.get(firstChar,2);// Get the first characters

  while(!fStream.eof() ){

    firstString = firstChar;

    //dump all comment lines else process them

    if(firstString == "#" ){
      fStream.getline(garbage,1000);
      fStream.get(firstChar,2);

    } else {
      if(!doneAlready){
        fStream >> m_tbHecPhiEuler >> m_tbHecThetaEuler >>m_tbHecPsiEuler >> m_tbHecXTrans >> m_tbHecYTrans >> m_tbHecZTrans;
        fStream >> m_tbEmecPhiEuler >> m_tbEmecThetaEuler >>m_tbEmecPsiEuler >> m_tbEmecXTrans >> m_tbEmecYTrans >> m_tbEmecZTrans;
        doneAlready = true;
      }
      fStream.getline(garbage,1000);// strip one more line to get to the end of file;
    }
  }
  fStream.close();
}


//====== testing and utility methods ======================

void Geometry::test(){

  map<int,float> testMap;

  testMap[41] = 26;
  testMap[113] = 42;
  testMap[197] = 900;
  testMap[102] = 400;
  testMap[135] = 600;
  draw2DHec(testMap,0,false);

  map<int,float> testMap2;
  testMap2[19] = 41;
  testMap2[12] =10;
  testMap2[1038] =80;
  testMap2[415] = 26;
  testMap2[571] =16;
  draw2DEmec(testMap2,0,true);

    draw2DEmecDetail(testMap2,0,true);
}

TMatrix* Geometry::constructRotationMatrix(double phi, double theta, double psi){

  Real_t rtheta[9];
  Real_t rphi[9];
  Real_t rpsi[9];

  rtheta[0]= cos(theta);
  rtheta[1]= 0;
  rtheta[2] = sin(theta);
  rtheta[3] = 0;
  rtheta[4] =1;
  rtheta[5] =0;
  rtheta[6] = -sin(theta);
  rtheta[7] = 0;
  rtheta[8] = cos(theta);

  rphi[0] = cos(phi);
  rphi[1] = sin(phi);
  rphi[2] = 0;
  rphi[3] = -sin(phi);
  rphi[4] = cos(phi);
  rphi[5] = 0;
  rphi[6] = 0;
  rphi[7] = 0;
  rphi[8] = 1;

  rpsi[0] = cos(psi);
  rpsi[1] = sin(psi);
  rpsi[2] = 0;
  rpsi[3] = -sin(psi);
  rpsi[4] = cos(psi);
  rpsi[5] = 0;
  rpsi[6] = 0;
  rpsi[7] = 0;
  rpsi[8] = 1;


  TMatrix rthetaMatrix(3,3,rtheta,"");
  TMatrix rphiMatrix(3,3,rphi,"");
  TMatrix rpsiMatrix(3,3,rpsi,"");

  TMatrix* final =  new TMatrix( rpsiMatrix*(rthetaMatrix*rphiMatrix) );

  return final;
}

void Geometry::histoEmecLayers(){
  vector<TH1F*> emecHisto;
  emecHisto.push_back(new TH1F("emec0","Emec Layer 0 Z",100,0,0));
  emecHisto.push_back(new TH1F("emec1","Emec Layer 1 Z",100,0,0));
  emecHisto.push_back(new TH1F("emec2","Emec Layer 2 Z",100,0,0));
  emecHisto.push_back(new TH1F("emec3","Emec Layer 3 Z",100,0,0));
  for(map<int,TVector3>::iterator it = m_emecLocalCenterMap.begin(); it != m_emecLocalCenterMap.end(); it++){
    emecHisto[m_emec_iz[it->first]]->Fill(it->second.z());
  }
  TCanvas* can = new TCanvas("emecCan1", "Emec Histograms",800,600);

  can->Divide(2,2);
  can->cd(1);
  emecHisto[0]->Draw();
  can->cd(2);
  emecHisto[1]->Draw();
  can->cd(3);
  emecHisto[2]->Draw();
  can->cd(4);
  emecHisto[3]->Draw();

  for(int i = 0; i < 4;i++){
    emecHisto[i]->GetXaxis()->SetTitle("Location (cm)");
    emecHisto[i]->GetYaxis()->SetTitle("frequency");
  }
}

void Geometry::drawJunk(){

  // This is example root code added in this class purely for illustrative purposses
  // it does not do anything useful.

  TGeoManager *geom = new TGeoManager("simple1", "Simple geometry");

  //--- define some materials
  TGeoMaterial *matVacuum = new TGeoMaterial("Vacuum", 0,0,0);
  TGeoMaterial *matAl = new TGeoMaterial("Al", 26.98,13,2.7);
  //--- define some media
  TGeoMedium *med;

  Double_t* something;
  TGeoMedium *Vacuum = new  TGeoMedium("Med1", 1, matVacuum);
  TGeoMedium *Al = new TGeoMedium("Med2", 2, matAl);

  //--- define the transformations
  TGeoTranslation *tr1 = new TGeoTranslation(20., 0, 0.);
  TGeoTranslation *tr2 = new TGeoTranslation(10., 0., 0.);
  TGeoTranslation *tr3 = new TGeoTranslation(10., 20., 0.);
  TGeoTranslation *tr4 = new TGeoTranslation(5., 10., 0.);
  TGeoTranslation *tr5 = new TGeoTranslation(20., 0., 0.);
  TGeoTranslation *tr6 = new TGeoTranslation(-5., 0., 0.);
  TGeoTranslation *tr7 = new TGeoTranslation(7.5, 7.5, 0.);
  TGeoRotation   *rot1 = new TGeoRotation("rot1", 90., 0., 90., 270., 0., 0.);
  TGeoCombiTrans *combi1 = new TGeoCombiTrans(7.5, -7.5, 0., rot1);
  TGeoTranslation *tr8 = new TGeoTranslation(7.5, -5., 0.);
  TGeoTranslation *tr9 = new TGeoTranslation(7.5, 20., 0.);
  TGeoTranslation *tr10 = new TGeoTranslation(85., 0., 0.);
  TGeoTranslation *tr11 = new TGeoTranslation(35., 0., 0.);
  TGeoTranslation *tr12 = new TGeoTranslation(-15., 0., 0.);
  TGeoTranslation *tr13 = new TGeoTranslation(-65., 0., 0.);
   TGeoTranslation  *tr14 = new TGeoTranslation(0,0,-100);
  TGeoCombiTrans *combi2 = new TGeoCombiTrans(0,0,100,
                                  new TGeoRotation("rot2",90,180,90,90,180,0));
  TGeoCombiTrans *combi3 = new TGeoCombiTrans(100,0,0,
                                  new TGeoRotation("rot3",90,270,0,0,90,180));
  TGeoCombiTrans *combi4 = new TGeoCombiTrans(-100,0,0,
                                  new TGeoRotation("rot4",90,90,0,0,90,0));
  TGeoCombiTrans *combi5 = new TGeoCombiTrans(0,100,0,
                                  new TGeoRotation("rot5",0,0,90,180,90,270));
  TGeoCombiTrans *combi6 = new TGeoCombiTrans(0,-100,0,
                                  new TGeoRotation("rot6",180,0,90,180,90,90));
   //--- make the top container volume
  Double_t worldx = 110.;
  Double_t worldy = 50.;
  Double_t worldz = 5.;
  TGeoVolume *top = geom->MakeBox("TOP", Vacuum, 270., 270., 120.);
  geom->SetTopVolume(top); // mandatory !
  //--- build other container volumes
  TGeoVolume *replica = geom->MakeBox("REPLICA", Vacuum,120,120,120);
  replica->SetVisibility(kFALSE);
  TGeoVolume *rootbox = geom->MakeBox("ROOT", Vacuum, 110., 50., 5.);
  rootbox->SetVisibility(kFALSE); // this will hold word 'ROOT'
   //--- make letter 'R'
  TGeoVolume *R = geom->MakeBox("R", Vacuum, 25., 25., 5.);
  R->SetVisibility(kFALSE);
  TGeoVolume *bar1 = geom->MakeBox("bar1", Al, 5., 25, 5.);
  bar1->SetLineColor(kRed);
  R->AddNode(bar1, 1, tr1);
  TGeoVolume *bar2 = geom->MakeBox("bar2", Al, 5., 5., 5.);
  bar2->SetLineColor(kRed);
  R->AddNode(bar2, 1, tr2);
  R->AddNode(bar2, 2, tr3);
  TGeoVolume *tub1 = geom->MakeTubs("tub1", Al, 5., 15., 5., 90., 270.);
  tub1->SetLineColor(kRed);
  R->AddNode(tub1, 1, tr4);
  TGeoVolume *bar3 = geom->MakeArb8("bar3", Al, 5.);
  bar3->SetLineColor(kRed);
  TGeoArb8 *arb = (TGeoArb8*)bar3->GetShape();
  arb->SetVertex(0, 15., -5.);
  arb->SetVertex(1, 5., -5.);
  arb->SetVertex(2, -10., -25.);
  arb->SetVertex(3, 0., -25.);
  arb->SetVertex(4, 15., -5.);
  arb->SetVertex(5, 5., -5.);
  arb->SetVertex(6, -10., -25.);
  arb->SetVertex(7, 0., -25.);
  R->AddNode(bar3, 1, gGeoIdentity);
   //--- make letter 'O'
  TGeoVolume *O = geom->MakeBox("O", Vacuum, 25., 25., 5.);
  O->SetVisibility(kFALSE);
  TGeoVolume *bar4 = geom->MakeBox("bar4", Al, 5., 7.5, 5.);
  bar4->SetLineColor(kYellow);
  O->AddNode(bar4, 1, tr5);
  O->AddNode(bar4, 2, tr6);
  TGeoVolume *tub2 = geom->MakeTubs("tub1", Al, 7.5, 17.5, 5., 0., 180.);
  tub2->SetLineColor(kYellow);
  O->AddNode(tub2, 1, tr7);
  O->AddNode(tub2, 2, combi1);
   //--- make letter 'T'
  TGeoVolume *T = geom->MakeBox("T", Vacuum, 25., 25., 5.);
  T->SetVisibility(kFALSE);
  TGeoVolume *bar5 = geom->MakeBox("bar5", Al, 5., 20., 5.);
  bar5->SetLineColor(kBlue);
  T->AddNode(bar5, 1, tr8);
  TGeoVolume *bar6 = geom->MakeBox("bar6", Al, 17.5, 5., 5.);
  bar6->SetLineColor(kBlue);
  T->AddNode(bar6, 1, tr9);
   //--- add letters to 'ROOT' container
  rootbox->AddNode(R, 1, tr10);
  rootbox->AddNode(O, 1, tr11);
  rootbox->AddNode(O, 2, tr12);
  rootbox->AddNode(T, 1, tr13);
   //--- add word 'ROOT' on each face of a cube
  replica->AddNode(rootbox, 1, tr14);
  replica->AddNode(rootbox, 2, combi2);
  replica->AddNode(rootbox, 3, combi3);
  replica->AddNode(rootbox, 4, combi4);
  replica->AddNode(rootbox, 5, combi5);
  replica->AddNode(rootbox, 6, combi6);
   //--- add four replicas of this cube to top volume
  top->AddNode(replica, 1, new TGeoTranslation(-150, -150, 0));
  top->AddNode(replica, 2, new TGeoTranslation(150, -150, 0));
  top->AddNode(replica, 3, new TGeoTranslation(150, 150, 0));
  top->AddNode(replica, 4, new TGeoTranslation(-150, 150, 0));
   //--- close the geometry
  geom->CloseGeometry();
   //--- draw the ROOT box
  geom->SetVisLevel(4);
  top->Draw();
  if (gPad) gPad->x3d();
}
